# -*- coding: utf-8 -*-
"""
Created on Tue Aug 15 11:41:20 2023

@author: ZHANG Jun
"""

# import important object.
from agat.app.cata.find_adsorption_site import *
from agat.app.cata.generate_adsorption_sites import *
from agat.app.cata.high_throughput_predict import *
from agat.app.cata.increase_vacuum import *
from agat.app.cata.volcano_2e import *
from agat.app.cata.volcano_plot import *




# -*- coding: utf-8 -*-
"""
Created on Sat Aug 12 16:08:41 2023

@author: ZHANG Jun
"""

import os

def generate_file_name(fname):
    while os.path.exists(fname):
        fname = fname + '_new'
    return fname



# -*- coding: utf-8 -*-
"""
Created on Tue Aug 15 11:41:20 2023

@author: ZHANG Jun
"""

from agat.lib.adsorbate_poscar import *
from agat.lib.AgatException import *
from agat.lib.file import *
from agat.lib.GatLib import *



# -*- coding: utf-8 -*-
"""
Created on Tue Jun 15 19:34:59 2021

@author: ZHANG Jun
"""

import tensorflow as tf
from scipy.stats import pearsonr
import os
import json
from dgl.data.utils import save_graphs
import multiprocessing
import time
import numpy as np
# from GatApp import GatApp # cannot import functions from each other

def config_parser(config):
    if isinstance(config, dict):
        return config
    elif isinstance(config, str):
        with open(config, 'r') as config_f:
            return json.load(config_f)
    elif isinstance(config, type(None)):
        return {}
    else:
        raise TypeError('Wrong configuration type.')

class EarlyStopping:
    def __init__(self, model, graph, logger, patience=10, folder='files'):
        self.model      = model
        self.patience   = patience
        self.counter    = 0
        self.best_score = None
        self.update     = None
        self.early_stop = False
        self.folder     = folder
        self._graph     = graph
        self.save_structure()



        # save a graph
        save_graphs(os.path.join(self.folder, 'graph_tmp.bin'), [graph],
                    {'prop': tf.constant([0.0])})

        self.logger     = logger

    def step(self, score, model):
        # score = mae
        if self.best_score is None:
            self.best_score = score
            self.update     = True
            self.save_checkpoint(model)
        elif score > self.best_score:
            self.counter += 1
            print(f'User log: EarlyStopping counter: {self.counter} out of {self.patience}',
                  file=self.logger)
            if self.counter >= self.patience:
                self.early_stop = True
        else:
            self.best_score = score
            self.update     = True
            self.save_checkpoint(model)
            self.counter = 0
        return self.early_stop

    def save_checkpoint(self, model):
        '''Saves model when validation loss decrease.'''
        model.save_weights(os.path.join(self.folder, 'gat.ckpt'))
        if model.batch_normalization:
            non_trainable = [x.numpy() for x in model.non_trainable_variables]
            with open(os.path.join(self.folder, 'Non_trainable_variables.npy'), 'wb') as f:
                for var in non_trainable:
                    np.save(f, var)

        print(f'User log: Save model weights with the best score: {self.best_score}',
              file=self.logger)

    def save_structure(self):
        gat_model = {}
        gat_model['Description']          = 'This is the architecture of a GAT model.'
        gat_model['num_gat_out_list']     = self.model.num_gat_out_list
        gat_model['num_readout_out_list'] = self.model.num_readout_out_list
        try:
            gat_model['head_list_en']     = self.model.head_list_en
        except AttributeError:
            pass
        try:
            gat_model['head_list_force']  = self.model.head_list_force
            gat_model['tail_readout_no_act'] = self.model.tail_readout_no_act
        except AttributeError:
            pass

        gat_model['att_activation']       = 'dgl.ops.edge_softmax'
        gat_model['embed_activation']     = self.model.embed_act_str
        gat_model['readout_activation']   = self.model.readout_act_str
        gat_model['bias']                 = self.model.bias
        gat_model['negative_slope']       = self.model.negative_slope
        gat_model['batch_normalization']  = self.model.batch_normalization

        node_depth                        = self._graph.ndata['h'].shape[1]
        gat_model['depth_of_node_representation'] = node_depth

        with open(os.path.join(self.folder, 'gat_model.json'), 'w') as f:
            json.dump(gat_model, f, indent=4)

def forward(model, graph): # , mean_prop=False): mean_prp: deprecated!
    """
    Parameters
    ----------
    model: a GAT model for crystal

    graph: a DGL graph

    mean_prop: whether to divide the graph output by the number of nodes.

    Return
    ----------
    `tf.reduce_mean(logit)`: `tf tensor`
        Prediction of the input `graph`
    or
    `tf.reduce_sum(logit)`: `tf tensor`
        Prediction of the input `graph`

    Note
    ----------
    The input graph will be modified by this function. So `clone()` it first.

    Modify this function when necessary.
        Using `tf.reduce_mean(logit)` when predict property per atom.

        Using `tf.reduce_sum(logit)` when predict the total property,
        for example, total energy.

    """
    features = graph.ndata['h']
    logit = model(features, graph)
    return tf.reduce_mean(logit)
    # if mean_prop:
    #     return tf.reduce_mean(logit)
    # else:
    #     return tf.reduce_sum(logit)

def forward_serial(model, graph, energy_list, forces_list, calculator): # clone graph before calling this function
    energy_per_atom, force_mat = calculator.get_force_energy_new(model, graph)
    energy_list.append(energy_per_atom)
    forces_list.append(force_mat)
    # print(energy_per_atom, energy_list)

def forward_parallel(model, graph_list, calculator):
    energy_list = multiprocessing.Manager().list()
    forces_list = multiprocessing.Manager().list()

    p           = multiprocessing.Pool(10)
    features = graph_list[1].ndata['h']
    p.apply_async(model, args=(features, graph_list[1],))
    # result2 = p.apply_async(model, args=(features, graph_list[1],))
    # print(result1.get())
    # p.apply_async(forward_serial, args=(model, graph_list[1], energy_list, forces_list, calculator,))
    # p.apply_async(forward_serial, args=(model, graph_list[2], energy_list, forces_list, calculator,))
    p.close()
    p.join()


    # energy_list, forces_list = [], []
    # # p           = multiprocessing.Pool(10)
    # with multiprocessing.Pool(processes = 10) as pool:
    #     # print('Here!!!')
    #     # for graph in graph_list:
    #         # forward_serial(model, graph, energy_list, forces_list, calculator)
    #         # p.apply_async(time.sleep, args=(1))
    #         # print(time.time())
    #         # args=(model, graph, energy_list, forces_list, calculator,)
    #         # print(type(args))
    #         # temp = pool.apply_async(func=forward_serial, args=(model, graph, energy_list, forces_list, calculator,),)
    #     temp = [pool.apply_async(func=forward_serial, args=(model, graph, energy_list, forces_list, calculator,),) for graph in graph_batch]
    #     # results = [t.get() for t in temp]
    #         # print(result)
    #         # result.get()
    #         # p.apply_async(forward_serial, args=(model, graph, energy_list, forces_list, calculator,))
    # # pool.close()
    # # pool.join()

    # # return list(energy_list), list(forces_list)
    # # return tf.convert_to_tensor(energy_list, dtype='float32'), tf.convert_to_tensor(forces_list, dtype='float32')
    # # return tf.convert_to_tensor(list(energy_list), dtype='float32'), tf.convert_to_tensor(list(forces_list), dtype='float32'), result
    # # return results

def load_gat_weights(model, graph, ckpt_path, logger, device): # clone graph before calling this function
    """
    Parameters
    ----------
    model: a fresh GAT model without trainable `variables`
    graph: a DGL graph
    ckpt_path: str
        Path to the saved checkpoint files

    Return
    ----------
    model: GAT model
        A GAT model with trainable `variables`
    """
    assert os.path.exists(ckpt_path + '.index'), "Checkpoint file not found when loading weights."
    # Varibles in model should be instantized and built first. See: https://github.com/tensorflow/tensorflow/issues/27937
    # forward(model, graph)    # build a gat model with varibles.
    # print(f'graph device: {graph.device}')
    # print(f'device passed into this function: {device}')

    graph = graph.to(device)
    # print(f'graph device: {graph.device}')
    with tf.device(device):
        model(graph)
    try:
        load_status = model.load_weights(ckpt_path)
        load_status.assert_consumed()  # check the load status
    except:
        print("User log: Weights detected but incompatible.", file=logger)

def accuracy(metrics, y_true, y_pred):
    metrics.update_state([y_true], [y_pred])
    return metrics.result().numpy()

def evaluate(metrics, model, g, y_true): # for now, this function is deprecated
    features = g.ndata['h']
    logits   = model(features, g)
    y_pred   = tf.reduce_mean(logits)
    return accuracy(metrics, y_true, y_pred)

def PearsonR(y_pred, y_true):
    return pearsonr(y_pred, y_true)[0]

def get_src_dst_data(graph):
    src, dst = graph.edges()
    src_data = tf.convert_to_tensor([graph.ndata['h'][x] for x in src])
    dst_data = tf.convert_to_tensor([graph.ndata['h'][x] for x in dst])
    return src_data, dst_data



# debug
if __name__ == '__main__':
    calculator = GatApp('1000_test\ckpt')
    graph_batch = [x.clone() for x in graph_list[0:10]]
    model = calculator.model
    start = time.time()
    # energy_list, forces_list, result = forward_parallel(model, graph_batch, calculator)
    forward_parallel(model, graph_batch, calculator)
    print(time.time() - start)



# -*- coding: utf-8 -*-
"""
Created on Thu Jun 17 20:41:00 2021

@author: ZHANG Jun
"""

import tensorflow as tf
from tensorflow.keras import layers
import numpy as np
from dgl import function as fn
from dgl.ops import edge_softmax

class GATLayer(layers.Layer):
    """
    Description:
    ----------
        Single graph attention network for crystal.
    Parameters
    ----------
    head_list: list
        A list of head mechanisms of this `GAT` layer.
    num_out: int
        Depth of node representation in the output of each head of this `GAT` layer.
    bias: Boolean
        Whether the dense layer uses a bias vector.
    negative_slope: float
        Negative slope coefficient
    activation: `tf.keras.activations`
        Activation function to use
    layer_type: str
        type of this gat layer.
        'input': input will be mapped into higher space.
        'hidden': dimension of feature representation will be kept.
        'output': output will be reduced into lower space.

    Meaning of abbreviations:
    ----------
    dist:
        distance matrix
    feat:
        features
    ft:
        features
    src:
        source node
    dst:
        destination node
    e:
        raw e_i_j: refer to: https://arxiv.org/abs/1710.10903
    a:
        alpha_i_j: refer to: https://arxiv.org/abs/1710.10903
    att:
        attention mechanism
	act:
        activation function
    """
    def __init__(self,
                 num_out,
                 num_heads,
                 bias=False,
                 negative_slope=0.2,
                 activation=None,
                 batch_normalization=False
                 ): # act_att='tanh'
        # The activation function of attention score on edges are deprecated.
        # See: https://docs.dgl.ai/en/0.6.x/generated/dgl.nn.functional.edge_softmax.html and https://www.tensorflow.org/api_docs/python/tf/keras/activations/softmax

        super(GATLayer, self).__init__()

        self.num_out        = num_out
        self.num_heads      = num_heads
        self.bias           = bias
        self.negative_slope = negative_slope
        self.activation     = activation
        # self.act_att        = act_att
        self.leaky_relu     = layers.LeakyReLU(alpha=negative_slope)
        self.xinit          = tf.keras.initializers.VarianceScaling(scale=np.sqrt(2),
                                                                     mode="fan_avg",
                                                                     distribution="untruncated_normal")
        self.batch_normalization = batch_normalization

    def build(self, input_shape):
        """
        The `build` method is thoroughly introduced by `TensorFlow`.
        Details: https://www.tensorflow.org/guide/keras/custom_layers_and_models
        """
        # print("Run build. ================") # for debug use
        self.w_att_src = tf.Variable(initial_value=self.xinit(shape=(1, self.num_heads, self.num_out),
                                                         dtype='float32'),
                                     trainable=True, name='w_att_src')
        self.w_att_dst = tf.Variable(initial_value=self.xinit(shape=(1, self.num_heads, self.num_out),
                                                         dtype='float32'),
                                     trainable=True, name='w_att_dst')

        self.dense = tf.keras.layers.Dense(self.num_out * self.num_heads,
                                                self.activation,
                                                self.bias,
                                                self.xinit,
                                                self.xinit)
        if self.batch_normalization:
            self.bn = tf.keras.layers.BatchNormalization()

    def call(self, feat, dist, graph): # feat with shape of (number of nodes, number of features of each node). The graph can have no values of nodes.
        """
        Description:
        ----------
            Forward this `GAT` layer.
        Parameters
        ----------
        feat: `tf.Tensor`
            Input features of all nodes (atoms).
        graph: `DGLGraph`
            A graph built with DGL.
        """

        # with graph.local_scope():
        feat_src = feat_dst = self.dense(feat)

        # attention
        feat_src = tf.reshape(feat_src, (-1, self.num_heads, self.num_out)) # shape: (number of nodes, number of heads, num out)
        feat_dst = tf.reshape(feat_dst, (-1, self.num_heads, self.num_out)) # shape: (number of nodes, number of heads, num out)

        # shape of e_src and e_dst after `tf.reduce_sum()`: (number of edges, number of heads, 1)
        # print(self.w_att_src)
        e_src = tf.reduce_sum(feat_src * self.w_att_src, axis=-1, keepdims=True)
        e_dst = tf.reduce_sum(feat_dst * self.w_att_dst, axis=-1, keepdims=True)

        # save on nodes
        graph.srcdata.update({'ft': feat_src, 'e_src': e_src})
        graph.dstdata.update({'e_dst': e_dst})

        # save on edges
        graph.apply_edges(fn.u_add_v('e_src', 'e_dst', 'e'))   # shape of e: (number of edges, number of heads, 1) # similar to the original paper, but feed to dense first, summation is the second.
        e = self.leaky_relu(graph.edata.pop('e'))  # shape of e: (number of edges, number of heads, 1)

        graph.edata['a']  = edge_softmax(graph, e) * dist # shape of a: (number of edges, number of heads, 1) # shape of dist: (number of edges, number of heads, 1)

        graph.update_all(fn.u_mul_e('ft', 'a', 'm'), fn.sum('m', 'ft')) # shape of ft: (number of nodes, number of heads, number of out)
        # print(tf.shape(graph.ndata['ft']))
        # print(tf.shape(graph.edata['a']))
        dst = tf.reshape(graph.ndata['ft'], (-1, self.num_heads * self.num_out)) # shape of `en`: (number of nodes, number of heads * number of out)
        if self.batch_normalization:
            dst = self.bn(dst)
        return dst # node_energy, node_force

# debug
if __name__ == '__main__':
    from ..data.data import CrystalGraph
    cg = CrystalGraph()
    bg = cg.get_graph('POSCAR')
    feat = bg.ndata['h']
    dist = bg.edata['dist']
    dist = tf.reshape(dist, (-1,1,1))

    SGL = GATLayer(50, 1)

    a = SGL(feat, dist, bg)

    # SGL1 = GATLayer(50, head_list=['mul', 'div', 'free'], layer_type='input')
    # bg = cg.get_graph('mp-1444.cif')
    # feat = bg.ndata['h']
    # SGL1(a, bg)

    # SGL2 = GATLayer(5, 50, layer_type='hidden')
    # bg = cg.get_graph('mp-1444.cif')
    # # feat = bg.ndata['h']
    # SGL2(a, bg)



# -*- coding: utf-8 -*-
"""
Created on Tue Aug 15 11:41:20 2023

@author: ZHANG Jun
"""

from agat.model.GatEnergyModel import *
from agat.model.GatForceModel import *
from agat.model.SingleGatLayer import *
from agat.model.ModelFit import *



# -*- coding: utf-8 -*-
"""
Created on Wed Apr 21 17:51:55 2021

@author: ZHANG Jun
"""

from pymatgen.core.periodic_table import Element
import numpy as np
import tensorflow as tf

all_elements = ['Ac', 'Ag', 'Al', 'Am', 'Ar', 'As', 'At', 'Au', 'B',  'Ba',
                'Be', 'Bh', 'Bi', 'Bk', 'Br', 'C',  'Ca', 'Cd', 'Ce', 'Cf',
                'Cl', 'Cm', 'Cn', 'Co', 'Cr', 'Cs', 'Cu', 'Db', 'Ds', 'Dy',
                'Er', 'Es', 'Eu', 'F',  'Fe', 'Fl', 'Fm', 'Fr', 'Ga', 'Gd',
                'Ge', 'H',  'He', 'Hf', 'Hg', 'Ho', 'Hs', 'I',  'In', 'Ir',
                'K',  'Kr', 'La', 'Li', 'Lr', 'Lu', 'Lv', 'Mc', 'Md', 'Mg',
                'Mn', 'Mo', 'Mt', 'N',  'Na', 'Nb', 'Nd', 'Ne', 'Nh', 'Ni',
                'No', 'Np', 'O',  'Og', 'Os', 'P',  'Pa', 'Pb', 'Pd', 'Pm',
                'Po', 'Pr', 'Pt', 'Pu', 'Ra', 'Rb', 'Re', 'Rf', 'Rg', 'Rh',
                'Rn', 'Ru', 'S',  'Sb', 'Sc', 'Se', 'Sg', 'Si', 'Sm', 'Sn',
                'Sr', 'Ta', 'Tb', 'Tc', 'Te', 'Th', 'Ti', 'Tl', 'Tm', 'Ts',
                'U',  'V',  'W',  'Xe', 'Y',  'Yb', 'Zn', 'Zr']

elements = ['Ac', 'Ag', 'Al', 'Am', 'As', 'Au', 'B',  'Ba', 'Be',
            'Bi', 'Br', 'C',  'Ca', 'Cd', 'Ce', 'Cl', 'Co', 'Cr', 'Cs',
            'Cu', 'Dy', 'Er', 'Eu', 'F',  'Fe', 'Ga', 'Gd', 'Ge', 'H',
            'Hf', 'Hg', 'Ho', 'I',  'In', 'Ir', 'K',  'La', 'Li', 'Lu',
            'Mg', 'Mn', 'Mo', 'N',  'Na', 'Nb', 'Nd', 'Ni', 'Np', 'O',
            'Os', 'P',  'Pb', 'Pd', 'Pm', 'Po', 'Pr', 'Pt', 'Pu', 'Ra',
            'Rb', 'Re', 'Rh', 'Ru', 'S',  'Sb', 'Sc', 'Se', 'Si', 'Sm',
            'Sn', 'Sr', 'Ta', 'Tb', 'Tc', 'Te', 'Th', 'Ti', 'Tl', 'Tm',
            'U',  'V',  'W',  'Y',  'Yb', 'Zn', 'Zr']

selected_elements = ['Ni', 'Co', 'Fe', 'Al', 'Ti']

# elements_not_used = ['Ar', 'Bh', 'Cn', 'Db', 'Ds', 'Fl', 'Hs', 'Lv', 'Mc', 'Mt', 'Nh', 'Og', 'Rf', 'Rg', 'Sg', 'Ts']

NiFe_ele = ['Fe', 'Ni']

Elements_used = elements

def get_atomic_features(Elements_used):
    """
    For more input features, please refer to: https://pymatgen.org/pymatgen.core.periodic_table.html

    Returns
    -------
    None.

    """
    atom_feat_data, atom_feat = [], {}
    for i, element in enumerate(Elements_used):
        atom_feat_tmp = []
        atom_feat_tmp.append(Element(element).Z)                # atomic number
        atom_feat_tmp.append(Element(element).atomic_radius)
        atom_feat_tmp.append(Element(element).molar_volume)
        atom_feat_tmp.append(Element(element).atomic_mass)
        atom_feat_tmp.append(Element(element).mendeleev_no)
        atom_feat_tmp.append(Element(element).X)                # Electronegativity of element
        atom_feat_tmp.append(Element(element).boiling_point)
        atom_feat_tmp.append(Element(element).melting_point)
        atom_feat_tmp.append(Element(element).row)
        atom_feat_tmp.append(Element(element).group)
        atom_feat_tmp.append(Element(element).max_oxidation_state)
        atom_feat_tmp.append(Element(element).min_oxidation_state)
        atom_feat_data.append(atom_feat_tmp)

    # scale the features
    atom_feat_data = np.array(atom_feat_data)
    max_list = np.max(atom_feat_data, axis=0)
    min_list = np.min(atom_feat_data, axis=0)

    for i in range(np.shape(atom_feat_data)[0]):
        for j in range(np.shape(atom_feat_data)[1]):
            atom_feat_data[i][j] = (atom_feat_data[i][j] - min_list[j]) / (max_list[j] - min_list[j])
    for i_index, i in enumerate(Elements_used):
        atom_feat[i] = atom_feat_data[i_index,:]
    return atom_feat

def get_atomic_feature_onehot(Elements_used):
    """
    Description
    ----------
    Get the atomic number of all considered elements.

    Returns
    -------
    atom_feat : dict
        atomic number of all considered elements
    """
    number_of_elements = len(Elements_used)

    atomic_number = [Element(x).Z for x in Elements_used]
    atoms_dict = dict(zip(atomic_number, Elements_used))

    atomic_number.sort()

    Elements_sorted = [atoms_dict[x] for x in atomic_number]

    keys               = Elements_sorted
    element_index      = [i for i, ele in enumerate(keys)]
    values             = tf.one_hot(element_index, number_of_elements)
    atom_feat          = dict(zip(keys, values))
    return atom_feat



"""
This is a modified DGL GAT model.
Inspired by: https://github.com/dmlc/dgl/blob/0.6.x/examples/tensorflow/gat/gat.py
"""

import tensorflow as tf
from tensorflow.keras import Model as tf_model
from agat.model.SingleGatLayer import GATLayer
from dgl.ops import edge_softmax

class EnergyGat(tf_model):
    """
    Description:
    ----------
        A GAT model with multiple gat layers.
    Parameters
    ----------
    num_gat_out_list: list
        A list of numbers that contains the representation dimension of each GAT layer.
    num_readout_out_list: list
        A list of numbers that contains the representation dimension of each readout layer.
    head_list_en: list
        A list contains the attention mechanisms of each head for the energy prediction.
    embed_activation: str
        TensorFlow activation function for embedding the inputs.
    readout_activation: str
        TensorFlow activation function for the readout layers.
    bias: boolean
        bias term of dense layers.
    negative_slope: float
        Negative slope of LeakyReLu function.
    Return of `call` method
    -----------------------
    en_h : tf Tensor
        Raw energy predictions of each atom (node).
    Important note
    -----------------------
    The last readout list must be one. Because the node energy or node force should have one value.
    """

    def __init__(self,
                 num_gat_out_list,
                 num_readout_out_list=[1],
                 head_list_en=['div'],
                 embed_activation='LeakyReLU',
                 readout_activation='LeakyReLU',
                 bias=True,
                 negative_slope=0.2):
        super(EnergyGat, self).__init__()

        # the dimension of input features will be determined automatically.
        self.num_gat_out_list     = num_gat_out_list
        self.num_gat_layer        = len(num_gat_out_list)
        self.num_readout_out_list = num_readout_out_list
        self.num_readout_layer    = len(num_readout_out_list)

        # embedding inputs in every gat layer, which is more flexible than embeding inputs here.

        self.head_list_en    = head_list_en
        self.num_heads_en    = len(head_list_en)
        self.negative_slope  = negative_slope
        self._activation_f   = {'LeakyReLU':    tf.keras.layers.LeakyReLU(alpha=self.negative_slope),
                                'relu':         tf.keras.activations.relu,
                                'tanh':         tf.keras.activations.tanh,
                                'none':         None,
                                'softmax':      tf.keras.activations.softmax,
                                'edge_softmax': edge_softmax}

        self.embed_act_str, self.readout_act_str = [embed_activation, readout_activation]
        self.embed_act       = self._activation_f[embed_activation]

        self.readout_act     = self._activation_f[readout_activation]
        self.bias            = bias

        self.head_fn         = {'mul' : self.mul,
                                'div' : self.div,
                                'free': self.free}

        self.gat_layers         = []
        self.en_read_out_layers = []
        self.batch_normalization = False

        # GAT layer
        for l in range(self.num_gat_layer):
            self.gat_layers.append(GATLayer(self.num_gat_out_list[l],
                                            self.num_heads_en,
                                            bias=self.bias,
                                            negative_slope=self.negative_slope,
                                            activation=self.embed_act))

        # energy readout layer
        for l in range(self.num_readout_layer-1):
            self.en_read_out_layers.append(tf.keras.layers.Dense(self.num_readout_out_list[l],
                                                                    self.readout_act,
                                                                    self.bias))
        self.en_read_out_layers.append(tf.keras.layers.Dense(self.num_readout_out_list[-1],
                                                                self._activation_f['none'],
                                                                self.bias))
    def mul(self, TfTensor):
        return TfTensor

    def div(self, TfTensor):
        return 1/TfTensor

    def free(self, TfTensor):
        return tf.constant(1.0, shape=TfTensor.shape)

    def get_head_mechanism(self, fn_list, TfTensor):
        """
        :param fn_list: A list of head mechanisms. For example: ['mul', 'div', 'free']
        :type fn_list: list
        :param TfTensor: A tensorflow tensor
        :type TfTensor: tf.tensor
        :return: A new tensor after the transformation.
        :rtype: tensor

        """
        TfTensor_list = []
        for func in fn_list:
            TfTensor_list.append(self.head_fn[func](TfTensor))
        return tf.concat(TfTensor_list, 1)

    def call(self, graph):
        '''
        Description:
        ----------
            `call` function of GAT model.
        Parameters
        ----------
        graph: `DGL.Graph`
            A graph.
        Return
        -----------------------
        en_h : tf Tensor
            Raw energy predictions of each atom (node).
        '''
        with graph.local_scope():
            h    = graph.ndata['h']                                    # shape: (number of nodes, dimension of one-hot code representation)
            dist = tf.reshape(graph.edata['dist'], (-1, 1, 1))         # shape: (number of edges, 1, 1)
            dist = tf.where(dist < 0.5, 0.5, dist)                     # This will creat a new `dist` variable, insted of modifying the original memory.
            dist = self.get_head_mechanism(self.head_list_en, dist)    # shape of dist: (number of edges, number of heads, 1)

            for l in range(self.num_gat_layer):
                h = self.gat_layers[l](h, dist, graph)                 # shape of h: (number of nodes, number of heads * num_out)

            # read-out layers
            for l in range(self.num_readout_layer):
                h = self.en_read_out_layers[l](h)
            return h

# debug
if __name__ == '__main__':
    model = GAT([30,40,50],
                 num_readout_out_list=[1],
                 head_list_en=['div', 'mul'],
                 embed_activation='LeakyReLU',
                 readout_activation='LeakyReLU',
                 bias=True,
                 negative_slope=0.2)

    from modules.Crystal2Graph import CrystalGraph
    cg = CrystalGraph()

    bg = cg.get_graph('POSCAR.txt', super_cell=False)
    en = model(bg).numpy()





# -*- coding: utf-8 -*-
"""
Created on Sun May 23 17:28:54 2021

@author: ZHANG Jun
"""

import numpy as np
import tensorflow as tf
import os
import time
import sys
import dgl

from agat.model.GatEnergyModel import EnergyGat
from agat.model.GatForceModel import ForceGat
# from modules.Crystal2Graph import ReadGraphs, TrainValTestSplit
from agat.lib.GatLib import EarlyStopping, PearsonR, load_gat_weights, config_parser # get_src_dst_data #, evaluate
from agat.default_parameters import default_train_config
from agat.data.data import ReadGraphs

class Train():
    def __init__(self, **train_config):
        self.train_config = {**default_train_config, **config_parser(train_config)}

        data_config = config_parser(os.path.join(self.train_config['dataset_path'], 'graph_build_scheme.json'))
        graph_reader = ReadGraphs(**{**data_config, **{'load_from_binary':True}})
        print(f'Reading graphs from: {self.train_config["dataset_path"]} ......')
        self.graph_list, self.graph_labels = graph_reader.read_all_graphs()

        # read previously subset index
        train_index, validation_index, test_index = (np.loadtxt(os.path.join(self.train_config['dataset_path'], 'train.txt'), dtype=int),
                                                     np.loadtxt(os.path.join(self.train_config['dataset_path'], 'validation.txt'), dtype=int),
                                                     np.loadtxt(os.path.join(self.train_config['dataset_path'], 'test.txt'), dtype=int))

        self.train_index, self.validation_index, self.test_index = train_index, validation_index, test_index
        self.graph_eg = self.graph_list[0]

        # # train dataset
        # self.validation_graphs = [self.graph_list[x] for x in self.validation_index]
        # self.validation_prop   = [self.graph_labels['energy_true'][x] for x in self.validation_index]

        # validation dataset
        self.validation_graphs = [self.graph_list[x] for x in self.validation_index]
        self.validation_prop   = [self.graph_labels['energy_true'][x] for x in self.validation_index]

        # test dataset
        self.test_graphs = [self.graph_list[x] for x in self.test_index]
        self.test_prop   = [self.graph_labels['energy_true'][x] for x in self.test_index]

        # check inputs
        self.train_config['batch_size'] = min(self.train_config['batch_size'],
                                              len(self.train_index),
                                              len(self.validation_index),
                                              len(self.test_index))
        self.train_config['validation_freq'] = min(self.train_config['validation_freq'],
                                                   len(self.train_index))
        self.train_config['validation_samples'] = min(self.train_config['validation_freq'],
                                                   len(self.validation_index))
        self.train_config['val_batch_size'] = min(self.train_config['val_batch_size'],
                                                  len(self.validation_index))

        # if self.train_config['load_graphs_on_gpu']:
        # # if True:
        #     self.graph_list = [x.to("/gpu:0") for x in self.graph_list]
        #     self.validation_graphs = [x.to("/gpu:0") for x in self.validation_graphs]
        #     self.test_graphs = [x.to("/gpu:0") for x in self.test_graphs]

        # prepare out file
        if not os.path.exists(self.train_config['output_files']):
            os.mkdir(self.train_config['output_files'])
        if not os.path.exists(os.path.join(self.train_config['output_files'], 'energy_ckpt')):
            os.mkdir(os.path.join(self.train_config['output_files'], 'energy_ckpt'))
        if not os.path.exists(os.path.join(self.train_config['output_files'], 'force_ckpt')):
            os.mkdir(os.path.join(self.train_config['output_files'], 'force_ckpt'))

    def fit_energy_model(self):
        # open log file
        energy_log = open('energy_train.log', 'w', buffering=1)

        # define computing device.
        if self.train_config['gpu_for_energy_train'] < 0:
            energy_model_device = "/cpu:0"
        else:
            energy_model_device = "/gpu:{}".format(self.train_config['gpu_for_energy_train'])
        print('User log: specified device for energy model:', energy_model_device, file=energy_log)

        # initialize a model
        energy_model = EnergyGat(self.train_config['energy_GAT_out_list'],
                     num_readout_out_list=self.train_config['energy_model_readout_list'],
                     head_list_en=self.train_config['energy_model_head_list'],
                     embed_activation=self.train_config['embed_activation'],
                     readout_activation=self.train_config['readout_activation'],
                     bias=self.train_config['bias'],
                     negative_slope=self.train_config['negative_slope'])

        # early stop
        if self.train_config['early_stop'] and self.train_config['train_energy_model']:
            stopper = EarlyStopping(energy_model, self.graph_eg, energy_log,
                                    patience=self.train_config['stop_patience'],
                                    folder=os.path.join(self.train_config['output_files'], 'energy_ckpt'))

        # detect the saved weights. Load them if possible.
        model_save_path = os.path.join(self.train_config['output_files'], 'energy_ckpt', 'gat.ckpt')
        print("User log: Model save path:", model_save_path, file=energy_log)
        if os.path.exists(model_save_path + '.index'):    # If the `checkpoint` file is saved before, try to load it.
            print("User log: Loading saved weights......", file=energy_log)
            g_init = self.graph_eg.clone() # Initialize trainable variables first.
            load_gat_weights(energy_model, g_init, model_save_path, energy_log, energy_model_device)
        else:
            print("User log: Weights not detected.", file=energy_log)

        if self.train_config['transfer_learning']:
            nontrainable_layers = [x for x in range(len(energy_model.layers))][1:self.train_config['trainable_layers']]
            for layer_i in nontrainable_layers:
                energy_model.layers[layer_i].trainable = False
# =============
        # start the training
        start_time = time.time()
        step       = 1

        # train
        flag = False
        for epoch in range(self.train_config['epochs']):
            np.random.shuffle(self.train_index) # shuffle training set every epoch

            # Split the data in batchs.
            train_index_batch = np.array_split(self.train_index, int(len(self.train_index) / self.train_config['batch_size'] + 1))
            en_pred_all, en_true_all = [], []

            for i_index, i in enumerate(train_index_batch):
                with tf.device("/cpu:0"):
                    g_batch = [self.graph_list[x] for x in i]
                    g_batch = dgl.batch(g_batch)
                    en_true = tf.stack([self.graph_labels['energy_true'][x] for x in i], axis=0)

                with tf.device(energy_model_device):
                    en_true = tf.identity(en_true)
                    g_batch = g_batch.to(energy_model_device)
                    with tf.GradientTape() as tape:
                        en_pred     = energy_model(g_batch)
                        batch_nodes = g_batch.batch_num_nodes()
                        en_pred     = tf.split(en_pred, batch_nodes)
                        en_pred     = tf.convert_to_tensor([tf.reduce_mean(x) for x in en_pred]) #tf.reduce_mean(en_pred, axis=1)

                        en_pred_all.append(en_pred)
                        en_true_all.append(en_true)
                        step += len(i)
                        loss_value = self.train_config['energy_loss_fcn'](en_true, en_pred) # calculate loss

                        # regularization. Specified in `Settings`
                        if self.train_config['L2_reg']:
                            for weight in energy_model.trainable_weights:
                                loss_value += self.train_config['weight_decay']*tf.nn.l2_loss(weight)
                    grads = tape.gradient(loss_value, energy_model.trainable_weights)
                    self.train_config['energy_optimizer'].apply_gradients(zip(grads, energy_model.trainable_weights))

                    batch_mae = self.train_config['mae'](en_true, en_pred)  # 不使用tf.keras.metrics.MeanAbsoluteError()。使用metrics，同样的输入，连续计算，输出不同。

                print("Epoch: {:0>4d} | Batch: {:0>4d} | Train_Loss (mse): {:1.4f} | Batch_MAE: {:1.4f} | Dur: {:.1f} s"
                      .format(epoch, i_index, loss_value, batch_mae, time.time() - start_time), file=energy_log)

                # evaluate every `validation_freq` epochs. `val` refers to validation.
                # `pred` refers to predict
                if step % self.train_config['validation_freq'] < self.train_config['batch_size'] and step > self.train_config['batch_size']:
                    print("========================================================================", file=energy_log)
                    en_val_true, en_val_pred = [], []

                    # with tf.device(energy_model_device):
                    validation_list = np.random.choice(range(len(self.validation_index)),
                                                       size=min(self.train_config['validation_samples'], len(self.validation_index)), replace=False)
                    validation_batch = np.array_split(validation_list, len(validation_list) / self.train_config['val_batch_size'])
                    for i in validation_batch:
                        with tf.device("/cpu:0"):
                            g_batch = [self.validation_graphs[x] for x in i]
                            g_batch = dgl.batch(g_batch)
                            en_true = tf.stack([self.validation_prop[x] for x in i], axis=0)

                        with tf.device(energy_model_device):
                            g_batch = g_batch.to(energy_model_device)
                            en_pred = energy_model(g_batch)
                            batch_nodes = g_batch.batch_num_nodes()
                            en_pred = tf.split(en_pred, batch_nodes)
                            en_pred = tf.convert_to_tensor([tf.reduce_mean(x) for x in en_pred])

                        with tf.device("/cpu:0"):
                            en_val_pred.append(tf.identity(en_pred))
                            en_val_true.append(en_true)

                    en_val_true, en_val_pred = tf.reshape(tf.concat(en_val_true, axis=0), shape=(-1)), tf.reshape(tf.concat(en_val_pred, axis=0), shape=(-1))
                    val_mae = self.train_config['mae'](en_val_true, en_val_pred)

                    dur = time.time() - start_time
                    en_true_all, en_pred_all = tf.reshape(tf.concat(en_true_all, axis=0), shape=(-1)), tf.reshape(tf.concat(en_pred_all, axis=0), shape=(-1))
                    print("Epoch: {:0>4d} | Train_MAE : {:1.4f} | Train_PearsonR: {:1.4f} | Val_MAE: {:1.4f} | Val_PearsonR: {:1.4f} | Dur: {:.1f} s".
                          format(epoch, self.train_config['mae'](en_true_all, en_pred_all),
                                 PearsonR(en_true_all, en_pred_all), val_mae,
                                 PearsonR(en_val_true, en_val_pred), dur),file=energy_log)

                    if self.train_config['early_stop'] and self.train_config['train_energy_model']:
                        if stopper.step(val_mae, energy_model):
                            print('User log: model summary:', file=energy_log)
                            energy_model.summary()
                            print('User log: model summary done.', file=energy_log)
                            flag = True
                            break
                        if stopper.update:
                            np.savetxt(os.path.join(self.train_config['output_files'], 'en_train_true.txt'), en_true_all, fmt='%.8f')    # if scale or mean prop is true, the saved data is not the original data
                            np.savetxt(os.path.join(self.train_config['output_files'], 'en_train_pred.txt'), en_pred_all, fmt='%.8f')    # if scale or mean prop is true, the saved data is not the original data
                            np.savetxt(os.path.join(self.train_config['output_files'], 'en_val_true.txt'),   en_val_true, fmt='%.8f')    # if scale or mean prop is true, the saved data is not the original data
                            np.savetxt(os.path.join(self.train_config['output_files'], 'en_val_pred.txt'),   en_val_pred, fmt='%.8f')    # if scale or mean prop is true, the saved data is not the original data
                        en_true_all, en_pred_all = [], []
                    else:
                        energy_model.save_weights(model_save_path)
                if flag:
                    break

        print('User log: model summary:', file=energy_log)
        energy_model.summary()
        print('User log: model summary done.', file=energy_log)
# =============
        # predict test set.
        y_true, y_pred = self.test_prop, [] # saved on GPU.
        # with tf.device(energy_model_device):
        batch_index = np.array_split(range(len(self.test_graphs)), len(self.test_graphs) / self.train_config['val_batch_size'])
        for i in batch_index:
            with tf.device("/cpu:0"):
                g_batch = [self.test_graphs[x] for x in i]
                g_batch = dgl.batch(g_batch)

            with tf.device(energy_model_device):
                g_batch = g_batch.to(energy_model_device)
                en_pred = energy_model(g_batch)
                batch_nodes = g_batch.batch_num_nodes()
                en_pred = tf.split(en_pred, batch_nodes)
                en_pred = tf.convert_to_tensor([tf.reduce_mean(x) for x in en_pred])

            with tf.device("/cpu:0"):
                y_pred.append(tf.identity(en_pred))
                mae_value = self.train_config['mae'](tf.convert_to_tensor(y_true), tf.concat(y_pred, axis=0))
        print("User log: predict MAE: {:.4f}".format(mae_value), file=energy_log)

        np.savetxt(os.path.join(self.train_config['output_files'], 'en_test_final_true.txt'), tf.convert_to_tensor(y_true).numpy(), fmt='%.8f')
        np.savetxt(os.path.join(self.train_config['output_files'], 'en_test_final_pred.txt'), tf.concat(y_pred, axis=0).numpy(), fmt='%.8f')

        energy_log.close()

    def fit_force_model(self):
        # open log file
        logf = open('force_train.log', 'w', buffering=1)

        # define computing device.
        if self.train_config['gpu_for_force_train'] < 0:
            device = "/cpu:0"
        else:
            device = "/gpu:{}".format(self.train_config['gpu_for_force_train'])
        print('User log: specified device for force model:', device, file=logf)

        model = ForceGat(self.train_config['force_GAT_out_list'],
                     num_readout_out_list=self.train_config['force_model_readout_list'],
                     head_list_force=self.train_config['force_model_head_list'],
                     embed_activation=self.train_config['embed_activation'],
                     readout_activation=self.train_config['readout_activation'],
                     bias=self.train_config['bias'],
                     negative_slope=self.train_config['negative_slope'],
                     batch_normalization=self.train_config['batch_normalization'],
                     tail_readout_no_act=self.train_config['tail_readout_noact'])

        # early stop
        if self.train_config['early_stop'] and self.train_config['train_force_model']:
            stopper = EarlyStopping(model, self.graph_eg, logf,
                                    patience=self.train_config['stop_patience'],
                                    folder=os.path.join(self.train_config['output_files'], 'force_ckpt'))

        # detect the saved weights. Load them if possible.
        model_save_path = os.path.join(self.train_config['output_files'], 'foce_ckpt', 'gat.ckpt')
        print("User log: Model save path:", model_save_path, file=logf)
        if os.path.exists(model_save_path + '.index'):    # If the `checkpoint` file is saved before, try to load it.
            print("User log: Loading saved weights......", file=logf)
            g_init = self.graph_eg.clone() # Initialize trainable variables first.
            load_gat_weights(model, g_init, model_save_path, logf, device)
            # if batch_normalization:
            #     with open(os.path.join(Project, CKPT, 'Non_trainable_variables.npy'), 'rb') as f:
            #         model.bn.moving_mean     = tf.Variable(np.load(f), trainable=False)
            #         model.bn.moving_variance = tf.Variable(np.load(f), trainable=False)
        else:
            print("User log: Weights not detected.", file=logf)

        if self.train_config['transfer_learning']:
            nontrainable_layers = [x for x in range(len(model.layers))][1:self.train_config['trainable_layers']]
            for layer_i in nontrainable_layers:
                model.layers[layer_i].trainable = False

        start_time     = time.time()
        step           = 1

        # train
        flag = False
        for epoch in range(self.train_config['epochs']):
            np.random.shuffle(self.train_index) # shuffle training set every epoch

            # Split the data in batchs.
            train_index_batch = np.array_split(self.train_index, int(len(self.train_index) / self.train_config['batch_size'] + 1))
            force_pred_all, force_true_all = [], []

            for i_index, i in enumerate(train_index_batch):
                with tf.device("/cpu:0"):
                    g_batch = [self.graph_list[x] for x in i]
                    g_batch = dgl.batch(g_batch)
                    force_true = g_batch.ndata['forces_true']

                with tf.device(device):
                    g_batch = g_batch.to(device)
                    force_true = tf.identity(force_true)
                    with tf.GradientTape() as tape:
                        force_pred = model(g_batch, training=True)

                        # mask fixed atoms. Comment the following lines for HECC.
                        ######################################################
                        if self.train_config['mask_fixed']:
                            constraints = tf.cast(g_batch.ndata['constraints'], dtype='bool')
                            force_pred  = tf.reshape(force_pred[constraints], [-1, 3]) # g.ndata['constraints'] can be considered as the `tf.boolean_mask`, the output will be flattened to a 1D tensor.
                            force_true  = tf.reshape(force_true[constraints], [-1, 3])
                        ######################################################

                        force_pred_all.append(force_pred)
                        force_true_all.append(force_true)
                        step += len(i)
                        if self.train_config['adsorbate']:
                            sample_weight = tf.where(tf.cast(g_batch.ndata['adsorbate'], dtype='bool'), self.train_config['adsorbate_coeff'], 1.0)
                            # sample_weight = tf.transpose(tf.stack([sample_weight, sample_weight, sample_weight], axis=0))
                            if self.train_config['mask_fixed']:
                                sample_weight = sample_weight[tf.math.reduce_any(constraints, axis=1)]
                        else:
                            sample_weight = None

                        loss_value = self.train_config['force_loss_fcn'](force_true, force_pred, sample_weight)

                        # regularization. Specified in `Settings`
                        if self.train_config['L2_reg']:
                            for weight in model.trainable_weights:
                                loss_value += self.train_config['weight_decay']*tf.nn.l2_loss(weight)
                    grads = tape.gradient(loss_value, model.trainable_weights)
                    self.train_config['force_optimizer'].apply_gradients(zip(grads, model.trainable_weights))

                    force_batch_mae = self.train_config['mae'](force_true, force_pred)  # 不使用tf.keras.metrics.MeanAbsoluteError()。若使用metrics，同样的输入，连续计算，输出不同。

                    print("Epoch: {:0>4d} | Batch: {:0>4d} | Loss (mse): {:1.4f} | MAE: {:1.4f} | Dur: {:.1f} s"
                          .format(epoch, i_index, loss_value, force_batch_mae, time.time() - start_time), file=logf)

                # evaluate every `validation_freq` epochs/graphs. `val` refers to validation.
                # `pred` refers to predict
                if step % self.train_config['validation_freq'] < self.train_config['batch_size'] and step > self.train_config['batch_size']:
                    print("========================================================================", file=logf)
                    force_val_true, force_val_pred = [], []
                    # with tf.device(device):
                    # save time if not forward all validation samples.
                    validation_list = np.random.choice(range(len(self.validation_index)),
                                                       size=min(self.train_config['validation_samples'], len(self.validation_index)), replace=False)
                    validation_batch = np.array_split(validation_list, len(validation_list) / self.train_config['val_batch_size'])
                    for i in validation_batch:
                        with tf.device("/cpu:0"):
                            g_batch    = [self.validation_graphs[x] for x in i]
                            g_batch    = dgl.batch(g_batch)
                            force_true = g_batch.ndata['forces_true']

                        with tf.device(device):
                            g_batch = g_batch.to(device)
                            force_pred = model(g_batch, training=False)

                            # mask fixed atoms. Comment the following lines for HECC.
                            ######################################################
                            if self.train_config['mask_fixed']:
                                constraints = tf.cast(g_batch.ndata['constraints'], dtype='bool')
                                force_pred  = force_pred[constraints] # g.ndata['constraints'] can be considered as the `tf.boolean_mask`, the output will be flattened to a 1D tensor.
                                force_true  = force_true[constraints]
                            ######################################################

                        with tf.device("/cpu:0"):
                            force_val_pred.append(tf.identity(force_pred))
                            force_val_true.append(force_true)

                    force_val_true, force_val_pred = tf.concat(force_val_true, axis=0), tf.concat(force_val_pred, axis=0)
                    force_val_mae = self.train_config['mae'](force_val_true, force_val_pred)

                    force_true_all, force_pred_all = tf.concat(force_true_all, axis=0), tf.concat(force_pred_all, axis=0)
                    dur = time.time() - start_time
                    print("Epoch: {:0>4d} | Train_Force_MAE : {:1.4f} | Train_Force_PearsonR: {:1.4f} | Force_Val_MAE: {:1.4f} | Force_Val_PearsonR: {:1.4f} | Dur: {:.1f} s".
                          format(epoch,
                                 self.train_config['mae'](force_true_all, force_pred_all), PearsonR(tf.reshape(force_true_all, [-1]), tf.reshape(force_pred_all, [-1])),
                                 force_val_mae, PearsonR(tf.reshape(force_val_true, [-1]), tf.reshape(force_val_pred, [-1])), dur),
                                 file=logf)

                    if self.train_config['early_stop'] and self.train_config['train_force_model']:
                        if stopper.step(force_val_mae, model):
                            print('User log: model summary:', file=logf)
                            model.summary()
                            print('User log: model summary done.', file=logf)
                            flag = True
                            break
                        if stopper.update:
                            np.savetxt(os.path.join(self.train_config['output_files'], 'force_train_true.txt'), tf.reshape(force_true_all, [-1]), fmt='%.8f')    # if scale or mean prop is true, the saved data is not the original data
                            np.savetxt(os.path.join(self.train_config['output_files'], 'force_train_pred.txt'), tf.reshape(force_pred_all, [-1]), fmt='%.8f')    # if scale or mean prop is true, the saved data is not the original data
                            np.savetxt(os.path.join(self.train_config['output_files'], 'force_val_true.txt'),   tf.reshape(force_val_true, [-1]), fmt='%.8f')    # if scale or mean prop is true, the saved data is not the original data
                            np.savetxt(os.path.join(self.train_config['output_files'], 'force_val_pred.txt'),   tf.reshape(force_val_pred, [-1]), fmt='%.8f')    # if scale or mean prop is true, the saved data is not the original data
                        force_true_all, force_pred_all = [], []

                    else:
                        model.save_weights(model_save_path)
                if flag:
                    break

        print('User log: model summary:', file=logf)
        model.summary()
        print('User log: model summary done.', file=logf)

        # predict test set.
        y_true, y_pred = [g.ndata['forces_true'] for g in self.test_graphs], [] # saved on GPU

        # with tf.device(device):
        batch_index = np.array_split(range(len(self.test_graphs)), len(self.test_graphs) / self.train_config['batch_size'])
        for i in batch_index:
            with tf.device("/cpu:0"):
                g_batch = [self.test_graphs[x] for x in i]
                g_batch = dgl.batch(g_batch)

            with tf.device(device):
                g_batch = g_batch.to(device)
                y_pred_tmp = model(g_batch)

            with tf.device("/cpu:0"):
                y_pred.append(tf.identity(y_pred_tmp))

        y_true, y_pred = tf.concat(y_true, axis=0), tf.concat(y_pred, axis=0)
        mae_value      = self.train_config['mae'](y_true, y_pred)

        print("User log: Final predict MAE of testset is: {:.4f}".format(mae_value), file=logf)
        np.savetxt(os.path.join(self.train_config['output_files'], 'force_test_final_true.txt'), tf.reshape(y_true, [-1]).numpy(), fmt='%.8f')
        np.savetxt(os.path.join(self.train_config['output_files'], 'force_test_final_pred.txt'), tf.reshape(y_pred, [-1]).numpy(), fmt='%.8f')
        logf.close()

if __name__ == '__main__':
    at = Train()
    at.fit_energy_model()
    at.fit_force_model()



# -*- coding: utf-8 -*-
"""
Created on Tue Jul  5 22:24:24 2022

@author: 18326
"""
import numpy as np

def volcano_2e():

    k = 8.617333262e-5 # unit: eV/K
    T = 300.0 # unit K
    k0 = 200.0 # unit: S^-1*site^-1
    vi = 21362619.98986784
    h = 4.1357 * 10**-15 # unit: eV * s
    vi = k*T/h  # This value will not affect the trend.

    # dissociative mechanism (2 e transfer)
    intervals = 201
    A_list, arg_list = np.zeros((intervals, intervals)), np.zeros((intervals, intervals))
    y_list = np.linspace(-2, 3.0, intervals)
    x_list = np.linspace(-1.5, 2.0, intervals)
    for i, O in enumerate(y_list):
        for j, OH in enumerate(x_list):

            g1 = 2.46 - 2 * 1.23
            g2 = O - 2 * 1.23
            g3 = OH - 1.23
            g4 = 0.0

            Ea = 1.8*O-2.95

            G1 = g2 - g1 # ΔG1
            G2 = g3 - g2 # ΔG2
            G3 = g4 - g3 # ΔG3

            eta = max([G1,G2,G3])

            k1 = vi*np.exp(-Ea/(k*T))
            k2 = vi*np.exp(-G1/(k*T))
            k3 = vi*np.exp(-G2/(k*T))
            k4 = vi*np.exp(-G3/(k*T))

            A1 = k*T*np.log10(k1/k0)
            A2 = k*T*np.log10(k2/k0)
            A3 = k*T*np.log10(k3/k0)
            A4 = k*T*np.log10(k4/k0)

            A = min(A1,A2,A3,A4)
            A_list[i,j] = A
            arg = np.argmax([A1,A2, A3])
            arg_list[i,j] = arg
    arg_list += 1

    A_list = np.vstack((x_list, A_list))
    A_list = np.hstack((np.vstack(([[0.0]],
                                     y_list.reshape(-1,1))),
                          A_list))
    arg_list = np.vstack((x_list, arg_list))
    arg_list = np.hstack((np.vstack(([[0.0]],
                                     y_list.reshape(-1,1))),
                          arg_list))

    np.savetxt('A_O_vs_OH_2e.txt', A_list, fmt='%f')
    np.savetxt('arg_O_vs_OH_2e.txt', arg_list, fmt='%f')

    # repeat the paper: J. Phys. Chem. B, Vol. 108, No. 46, 2004
    intervals = 201
    A_list, arg_list = np.zeros((intervals, intervals)), np.zeros((intervals, intervals))
    y_list = np.linspace(-2, 3.0, intervals)
    x_list = np.linspace(-1.5, 2.0, intervals)
    for i, O in enumerate(y_list):
        for j, OH in enumerate(x_list):

            Ea = 1.8*O-2.89    # From table 2

            G0 = O - 2.45      # From table 2
            G1 = OH - O + 0.97 # ΔG1 From table 2
            G2 = -OH + 1.48    # ΔG2 From table 2

            k1 = vi*np.exp(-Ea/(k*T))
            k2 = vi*np.exp(-G1/(k*T))
            k3 = vi*np.exp(-G2/(k*T))

            A1 = k*T*np.log10(k1/k0)
            A2 = k*T*np.log10(k2/k0)
            A3 = k*T*np.log10(k3/k0)

            A = min(A1,A2,A3)
            A_list[i,j] = A

    A_list = np.vstack((x_list, A_list))
    A_list = np.hstack((np.vstack(([[0.0]],
                                     y_list.reshape(-1,1))),
                          A_list))

    np.savetxt('A_O_vs_OH_2e.txt', A_list, fmt='%f')



"""
This is a modified DGL GAT model.
Inspired by: https://github.com/dmlc/dgl/blob/0.6.x/examples/tensorflow/gat/gat.py
"""

import tensorflow as tf
from tensorflow.keras import Model as tf_model
from agat.model.SingleGatLayer import GATLayer
from dgl.ops import edge_softmax
from dgl import function as fn

class ForceGat(tf_model):
    """
    Description:
    ----------
        A GAT model with multiple gat layers.
    Parameters
    ----------
    num_gat_out_list: list
        A list of numbers that contains the representation dimension of each GAT layer.
    num_readout_out_list: list
        A list of numbers that contains the representation dimension of each readout layer.
    head_list_en: list
        A list contains the attention mechanisms of each head for the energy prediction.
    head_list_force: list
        A list contains the attention mechanisms of each head for the force prediction.
    att_activation: str
        TensorFlow activation function for calculating the attention score `a`.
    embed_activation: str
        TensorFlow activation function for embedding the inputs.
    readout_activation: str
        TensorFlow activation function for the readout layers.
    bias: boolean
        bias term of dense layers.
    negative_slope: float
        Negative slope of LeakyReLu function.
    Return of `call` method
    -----------------------
    en_h : tf Tensor
        Raw energy predictions of each atom (node).
    graph.ndata['force_pred'] : tf Tensor
        Raw force predictions of each atom (node).
    Important note
    -----------------------
    The last readout list must be one. Because the node energy or node force should have one value.
    The Gaussian expansion is not well tested and should be deprecated.
    """

    def __init__(self,
                 num_gat_out_list,
                 num_readout_out_list=[1],
                 head_list_force=['div'],
                 embed_activation='LeakyReLU',
                 readout_activation='LeakyReLU',
                 bias=True,
                 negative_slope=0.2,
                 batch_normalization=False,
                 tail_readout_no_act=1):
        super(ForceGat, self).__init__()

        # the dimension of input features will be determined automatically.
        self.num_gat_out_list     = num_gat_out_list
        self.num_gat_layer        = len(num_gat_out_list)
        self.num_readout_out_list = num_readout_out_list
        self.num_readout_layer    = len(num_readout_out_list)

        # embedding inputs in every gat layer, which is more flexible than embeding inputs here.

        self.head_list_force = head_list_force
        self.num_heads_force = len(head_list_force)
        self.negative_slope  = negative_slope
        self._activation_f   = {'LeakyReLU':    tf.keras.layers.LeakyReLU(alpha=self.negative_slope),
                                'relu':         tf.keras.activations.relu,
                                'elu':          tf.keras.activations.elu,
                                'tanh':         tf.keras.activations.tanh,
                                'none':         None,
                                'softmax':      tf.keras.activations.softmax,
                                'edge_softmax': edge_softmax}

        self.embed_act_str, self.readout_act_str = [embed_activation, readout_activation]
        self.embed_act       = self._activation_f[embed_activation]

        self.readout_act     = self._activation_f[readout_activation]
        self.bias            = bias

        self.head_fn         = {'mul' : self.mul,
                                'div' : self.div,
                                'free': self.free}

        self.gat_layers            = []
        self.force_read_out_layers = []

        self.batch_normalization   = batch_normalization

        # self.Gaussian_expansion    = Gaussian_expansion
        # self.dist_miu              = dist_miu
        # self.dist_std              = dist_std

        self.tail_readout_no_act   = tail_readout_no_act

        # GAT layer
        for l in range(self.num_gat_layer):
            self.gat_layers.append(GATLayer(self.num_gat_out_list[l],
                                            self.num_heads_force,
                                            bias=self.bias,
                                            negative_slope=self.negative_slope,
                                            activation=self.embed_act,
                                            batch_normalization=self.batch_normalization))

        if self.batch_normalization:
            self.bn = tf.keras.layers.BatchNormalization()

        # energy readout layer
        for l in range(self.num_readout_layer-self.tail_readout_no_act):
            self.force_read_out_layers.append(tf.keras.layers.Dense(self.num_readout_out_list[l],
                                                                    self.readout_act,
                                                                    self.bias))
        for l in range(self.tail_readout_no_act):
            self.force_read_out_layers.append(tf.keras.layers.Dense(self.num_readout_out_list[l-self.tail_readout_no_act],
                                                                    self._activation_f['none'],
                                                                    self.bias))

    def mul(self, TfTensor):
        return TfTensor

    def div(self, TfTensor):
        return 1/TfTensor

    def free(self, TfTensor):
        return tf.constant(1.0, shape=TfTensor.shape)

    def get_head_mechanism(self, fn_list, TfTensor):
        """
        :param fn_list: A list of head mechanisms. For example: ['mul', 'div', 'free']
        :type fn_list: list
        :param TfTensor: A tensorflow tensor
        :type TfTensor: tf.tensor
        :return: A new tensor after the transformation.
        :rtype: tensor

        """
        TfTensor_list = []
        for func in fn_list:
            TfTensor_list.append(self.head_fn[func](TfTensor))
        return tf.concat(TfTensor_list, 1)

    def call(self, graph): #, Training=None, moving_mean=None, moving_var=None):
        '''
        Description:
        ----------
            `call` function of GAT model.
        Parameters
        ----------
        graph: `DGL.Graph`
            A graph.
        Return
        -----------------------
        en_h : tf Tensor
            Raw energy predictions of each atom (node).
        graph.ndata['force_pred'] : tf Tensor
            Raw force predictions of each atom (node).
        '''
        with graph.local_scope():
            h    = graph.ndata['h']                                    # shape: (number of nodes, dimension of one-hot code representation)
            dist = tf.reshape(graph.edata['dist'], (-1, 1, 1))         # shape: (number of edges, 1, 1)
            # if self.Gaussian_expansion:
            #     dist = tf.math.exp(-(dist - self.dist_miu) / self.dist_std)
            #     # dist = tf.where(dist < self.dist_miu, dist * -1 + 1, dist * + 1)
            # else:
            #     dist = tf.where(dist < 0.5, 0.5, dist)                 # This will creat a new `dist` variable, insted of modifying the original memory.
            dist = tf.where(dist < 0.5, 0.5, dist)                 # This will creat a new `dist` variable, insted of modifying the original memory.
            dist = self.get_head_mechanism(self.head_list_force, dist) # shape of dist: (number of edges, number of heads, 1)

            for l in range(self.num_gat_layer):
                h = self.gat_layers[l](h, dist, graph)                 # shape of h: (number of nodes, number of heads * num_out)

            # Predict force in real space.
            graph.ndata['node_force']   = h

            graph.apply_edges(fn.u_add_v('node_force', 'node_force', 'score'))    #!!!             # shape of score: (number of edges, ***, 1)

            score = tf.reshape(graph.edata['score'],(-1, self.num_heads_force, self.num_gat_out_list[-1])) / dist
            score = tf.reshape(score, (-1, self.num_heads_force * self.num_gat_out_list[-1]))
            if self.batch_normalization:
                score            = self.bn(score)
            # self.moving_mean = self.bn.moving_mean
            # self.moving_var  = self.bn.moving_variance

            for l in range(self.num_readout_layer):
                score = self.force_read_out_layers[l](score)

            graph.edata['score_vector'] = score * graph.edata['direction']      # shape (number of edges, 1)

            graph.update_all(fn.copy_e('score_vector', 'm'), fn.sum('m', 'force_pred'))        # shape of graph.ndata['force_pred']: (number of nodes, 3)
            return graph.ndata['force_pred']

# debug
if __name__ == '__main__':
    model = GAT([30,40,50],
                 num_readout_out_list=[1],
                 head_list_force=['div', 'mul'],
                 embed_activation='LeakyReLU',
                 readout_activation='LeakyReLU',
                 bias=True,
                 negative_slope=0.2)

    from modules.Crystal2Graph import CrystalGraph
    cg = CrystalGraph()

    bg = cg.get_graph('POSCAR.txt', super_cell=False)
    force = model(bg).numpy()



# -*- coding: utf-8 -*-
"""
Created on Fri Sep  3 17:16:24 2021

@author: ZHANG Jun
"""

import json
import os
import sys
import tensorflow as tf
from agat.data.data import CrystalGraph
from agat.model.GatEnergyModel import EnergyGat as GATE
from agat.model.GatForceModel import ForceGat as GATF
from dgl.data.utils import load_graphs
from ase import Atoms
from ase.optimize import BFGS
from ase.calculators.calculator import Calculator
from ase.io import read, write

class GatApp(object):
    """ Basic application of well-trained AGAT model.

    :param energy_model_save_path: Directory for the saved energy model.
    :type energy_model_save_path: str
    :param force_model_save_path: Directory for the saved force model
    :type force_model_save_path: str
    :param force_model_save_path: Directory for the json file of how to build graphs. This file is normally saved when you build your dataset.
    :type force_model_save_path: str
    :param load_model: Load well-trained models or not, defaults to True
    :type load_model: bool, optional
    :param gpu: Predict with GPU or CPU, negative value for GPU, defaults to -1
    :type gpu: int, optional
    :return: Energy and forces
    :rtype: tensorflow EagerTensor.

    Example::

        energy_model_save_dir = os.path.join('out_file', 'energy_ckpt')
        force_model_save_dir = os.path.join('out_file', 'force_ckpt')
        graph_build_scheme_dir = 'dataset'
        app = GatApp(energy_model_save_dir, force_model_save_dir, graph_build_scheme_dir)

        graph, info = app.get_graph('POSCAR')
        energy = app.get_energy(graph)
        forces = app.get_forces(graph)

    """
    def __init__(self, energy_model_save_path, force_model_save_path, graph_build_scheme_path, load_model=True, gpu=-1):

        self.energy_model_save_path = energy_model_save_path
        self.force_model_save_path  = force_model_save_path
        self.graph_build_scheme_path = graph_build_scheme_path
        self.gpu                    = gpu
        if gpu < 0:
            self.device             = "/cpu:0"
        else:
            self.device             = "/gpu:{}".format(gpu)

        if load_model:
            self.energy_model       = self.load_energy_model(energy_model_save_path)
            self.force_model        = self.load_force_model(force_model_save_path)

        self.graph_build_scheme     = self.load_graph_build_scheme(self.graph_build_scheme_path) # you can also load it from `force_model_save_path`

        build_properties = {'energy': False, 'forces': False, 'cell': False,
                            'cart_coords': False, 'frac_coords': False, 'path': False}
        self.graph_build_scheme['build_properties'] = {**self.graph_build_scheme['build_properties'], **build_properties}
        self.cg                     = CrystalGraph(**self.graph_build_scheme)

    def load_graph_build_scheme(self, path):
        """ Load graph building scheme. This file is normally saved when you build your dataset.

        :param path: Directory for storing ``graph_build_scheme.json`` file.
        :type path: str
        :return: A dict denotes how to build the graph.
        :rtype: dict

        """

        json_file  = os.path.join(path, 'graph_build_scheme.json')
        assert os.path.exists(json_file), f"{json_file} file dose not exist."
        with open(json_file, 'r') as jsonf:
            graph_build_scheme = json.load(jsonf)
        return graph_build_scheme

    def load_energy_model(self, energy_model_save_path):
        """ Load the energy model.

        :param energy_model_save_path: Directory for the saved energy model.
        :type energy_model_save_path: str
        :return: An AGAT model
        :rtype: agat.model.GatEnergyModel.EnergyGat

        """

        json_file  = os.path.join(energy_model_save_path, 'gat_model.json')
        graph_file = os.path.join(energy_model_save_path, 'graph_tmp.bin')
        ckpt_file  = os.path.join(energy_model_save_path, 'gat.ckpt')

        for f in [json_file, graph_file, ckpt_file + '.index']:
            assert os.path.exists(f), f"{f} file dose not exist."

        # load json file
        with open(json_file, 'r') as jsonf:
            model_config = json.load(jsonf)

        # build a model
        model =  GATE(model_config['num_gat_out_list'],
                      num_readout_out_list = model_config['num_readout_out_list'],
                      head_list_en         = model_config['head_list_en'],
                      embed_activation     = model_config['embed_activation'],
                      readout_activation   = model_config['readout_activation'],
                      bias                 = model_config['bias'],
                      negative_slope       = model_config['negative_slope'])

        # load weights
        graph_tmp, label_tmp = load_graphs(graph_file)
        graph_tmp = graph_tmp[0].to(self.device)
        with tf.device(self.device):
            model(graph_tmp)
        load_status          = model.load_weights(ckpt_file)
        load_status.assert_consumed()
        print(f'Load energy model weights from {ckpt_file} successfully.')
        return model

    def load_force_model(self, force_model_save_path):
        """ Load the force model.

        :param force_model_save_path: Directory for the saved force model.
        :type force_model_save_path: str
        :return: An AGAT model
        :rtype: agat.model.GatForceModel.ForceGat

        """
        json_file  = os.path.join(force_model_save_path, 'gat_model.json')
        graph_file = os.path.join(force_model_save_path, 'graph_tmp.bin')
        ckpt_file  = os.path.join(force_model_save_path, 'gat.ckpt')

        for f in [json_file, graph_file, ckpt_file + '.index']:
            assert os.path.exists(f), f"{f} file dose not exist."

        # load json file
        with open(json_file, 'r') as jsonf:
            model_config = json.load(jsonf)

        # build a model
        model =  GATF(model_config['num_gat_out_list'],
                      model_config['num_readout_out_list'],
                      model_config['head_list_force'],
                      model_config['embed_activation'],
                      model_config['readout_activation'],
                      model_config['bias'],
                      model_config['negative_slope'],
                      model_config['batch_normalization'],
                      model_config['tail_readout_no_act'])

        # load weights
        graph_tmp, label_tmp = load_graphs(graph_file)
        graph_tmp = graph_tmp[0].to(self.device)
        with tf.device(self.device):
            model(graph_tmp)
        load_status          = model.load_weights(ckpt_file)
        load_status.assert_consumed()
        print(f'Load force model weights from {ckpt_file} successfully.')
        return model

    def get_graph(self, fname):
        """ Build agat graph from VASP file.

        :param fname: The file name.
        :type fname: str
        :return: dgl graph.
        :rtype: dgl.heterograph.DGLGraph

        """

        return self.cg.get_graph(fname) # No need to read True forces for application

    def get_energy(self, graph):
        """ Forward the energy model.

        :param graph: dgl graph.
        :type graph: dgl.heterograph.DGLGraph
        :return: Total energy.
        :rtype: tf EagerTensor.

        """

        with tf.device(self.device):
            # graph.to(self.device)
            energy = tf.reduce_sum(self.energy_model(graph)).numpy()
        return energy

    def get_energies(self, graph):
        """ Forward the energy model.

        :param graph: dgl graph.
        :type graph: dgl.heterograph.DGLGraph
        :return: Energies of all atoms.
        :rtype: numpy.ndarray.

        """

        with tf.device(self.device):
            # graph.to(self.device)
            return self.energy_model(graph).numpy()

    def get_forces(self, graph):
        """ Forward the force model.

        :param graph: dgl graph.
        :type graph: dgl.heterograph.DGLGraph
        :return: Atomic forces.
        :rtype: numpy.ndarray.

        """
        with tf.device(self.device):
            # graph.to(self.device)
            return self.force_model(graph).numpy()

    def get_stress(self):
        """

        :return: DESCRIPTION
        :rtype: TYPE

        """

        return None

class GatAseCalculator(Calculator):
    """ Calculator with ASE module.

    :param energy_model_save_path: Directory for the saved energy model.
    :type energy_model_save_path: str
    :param force_model_save_path: Directory for the saved force model.
    :type force_model_save_path: str
    :param graph_build_scheme_path: Directory for the json file of how to build graphs. This file is normally saved when you build your dataset.
    :type graph_build_scheme_path: str
    :param load_model: Load well-trained models or not, defaults to True
    :type load_model: bool, optional
    :param gpu: Predict with GPU or CPU, negative value for GPU, defaults to -1
    :type gpu: int, optional
    :param **kwargs: kwargs
    :type **kwargs: dict/kwargs


    Example::

        energy_model_save_dir = os.path.join('out_file', 'energy_ckpt')
        force_model_save_dir = os.path.join('out_file', 'force_ckpt')
        graph_build_scheme_dir = 'dataset'
        poscar = read('POSCAR')
        calculator=GatAseCalculator(energy_model_save_dir,
                                    force_model_save_dir,
                                    graph_build_scheme_dir)
        poscar = Atoms(poscar, calculator=calculator)
        dyn = BFGS(poscar, trajectory='test.traj')
        dyn.run(fmax=0.005)

        traj = read('test.traj', index=':')
        write("XDATCAR.gat", traj)

    """
    implemented_properties = ['energy', 'energies', 'free_energy', 'forces', 'stress', 'stresses']
    default_parameters     = { }
    def __init__(self,
                 energy_model_save_path,
                 force_model_save_path,
                 graph_build_scheme_path,
                 load_model=True,
                 gpu       = -1,
                 **kwargs):

        Calculator.__init__(self, **kwargs)
        self.app = GatApp(energy_model_save_path,
                          force_model_save_path,
                          graph_build_scheme_path,
                          load_model=True,
                          gpu=gpu)

    def calculate(self, atoms=None, properties=None, system_changes=['positions', 'numbers', 'cell', 'pbc']):
        if properties is None:
            properties = self.implemented_properties

        # read graph
        graph, info = self.app.get_graph(atoms)
        # print(graph.device)
        graph = graph.to(self.app.device)
        # print(graph.device)

        # get results
        energy = self.app.get_energy(graph)
        forces = self.app.get_forces(graph)

        self.results = {'energy': energy,
                        'forces': forces} # ,                        'energies': energies

class GatCalculator(Calculator):
    """ Calculator with ASE module. Pymatgen is also needed.

    :param energy_model_save_path: Directory for the saved energy model.
    :type energy_model_save_path: str
    :param force_model_save_path: Directory for the saved force model.
    :type force_model_save_path: str
    :param graph_build_scheme_path: Directory for the json file of how to build graphs. This file is normally saved when you build your dataset.
    :type graph_build_scheme_path: str
    :param load_model: Load well-trained models or not, defaults to True
    :type load_model: bool, optional
    :param **kwargs: kwargs
    :type **kwargs: kwargs

    Example::

        energy_model_save_dir = os.path.join('out_file', 'energy_ckpt')
        force_model_save_dir = os.path.join('out_file', 'force_ckpt')
        graph_build_scheme_dir = 'dataset'
        poscar = read('POSCAR')
        calculator=GatCalculator(energy_model_save_dir,
                                    force_model_save_dir,
                                    graph_build_scheme_dir)
        poscar = Atoms(poscar, calculator=calculator)
        dyn = BFGS(poscar, trajectory='test.traj')
        dyn.run(fmax=0.005)

        traj = read('test.traj', index=':')
        write("XDATCAR.gat", traj)

    """
    implemented_properties = ['energy', 'energies', 'free_energy', 'forces', 'stress', 'stresses']
    default_parameters     = { }
    def __init__(self,
                 energy_model_save_path,
                 force_model_save_path,
                 graph_build_scheme_path,
                 load_model=True, **kwargs):


        Calculator.__init__(self, **kwargs)
        self.app = GatApp(energy_model_save_path,
                          force_model_save_path,
                          graph_build_scheme_path,
                          load_model=True)


    def calculate(self, atoms=None, properties=None, system_changes=['positions', 'numbers', 'cell', 'pbc']):
        if properties is None:
            properties = self.implemented_properties

        # save `atom` object to a file or to the memory
        write('CONTCAR.gat', atoms)

        # read file with MP
        graph, info = self.app.get_graph('CONTCAR.gat')
        # graph = graph.to(self.app.device)
        # print(graph.device)

        # get results
        energy = self.app.get_energy(graph)
        forces = self.app.get_forces(graph)

        self.results = {'energy': energy,
                        'forces': forces} # ,                        'energies': energies

# debug
if __name__ == '__main__':
    energy_model_save_dir = os.path.join('out_file', 'energy_ckpt')
    force_model_save_dir = os.path.join('out_file', 'force_ckpt')
    graph_build_scheme_dir = 'dataset'
    app = GatApp(energy_model_save_dir, force_model_save_dir, graph_build_scheme_dir)

    import time
    start = time.time()
    for i in range(10):
        graph, info = app.get_graph('POSCAR')
        print(time.time() - start)

    import time
    start = time.time()
    for i in range(10):
        energy = app.get_energy(graph)
        forces = app.get_forces(graph)
        print(time.time() - start)

    #BFGS
    poscar = read('POSCAR')
    calculator=GatAseCalculator(energy_model_save_dir,
                                force_model_save_dir,
                                graph_build_scheme_dir)
    poscar = Atoms(poscar, calculator=calculator)
    dyn = BFGS(poscar, trajectory='test.traj')
    dyn.run(fmax=0.005)

    traj = read('test.traj', index=':')
    write("XDATCAR.gat", traj)

    # with pymatgen
    poscar = read('POSCAR')
    calculator=GatCalculator(energy_model_save_dir,
                                force_model_save_dir,
                                graph_build_scheme_dir)
    poscar = Atoms(poscar, calculator=calculator)
    dyn = BFGS(poscar, trajectory='test.traj')
    dyn.run(fmax=0.005)

    traj = read('test.traj', index=':')
    write("XDATCAR.gat", traj)

    # from ase.optimize import BFGS, LBFGS, LBFGSLineSearch, GPMin, FIRE
    # import time
    # from ase.optimize.sciopt import SciPyFminBFGS, SciPyFminCG

    # for optimizer in (BFGS, LBFGS, LBFGSLineSearch, GPMin, FIRE):
    #     start = time.time()
    #     poscar = read('POSCAR')
    #     calculator=GatAseCalculator(energy_model_save_dir,
    #                                 force_model_save_dir,
    #                                 graph_build_scheme_dir)
    #     poscar = Atoms(poscar, calculator=calculator)
    #     # dyn = optimizer(poscar, trajectory='test.traj', maxstep=0.05)
    #     dyn = optimizer(poscar, trajectory='test.traj', callback_always=True)
    #     dyn.run(fmax=0.05, steps =5)
    #     print(f'Time used for {optimizer} is {time.time()-start}')





# -*- coding: utf-8 -*-
"""
Created on Tue Aug 15 11:41:20 2023

@author: ZHANG Jun
"""

__version__ = '7.12'

import os
os.environ['DGLBACKEND']="tensorflow"

# import important object.
from agat.data.data import AgatDatabase
from agat.model.ModelFit import Train
from agat.app.cata import high_throughput_predict



# -*- coding: utf-8 -*-
"""
Created on Tue Aug 15 11:41:20 2023

@author: ZHANG Jun
"""

def main(formula):
    from agat.data.data import AgatDatabase
    from agat.model.ModelFit import Train
    from agat.app.cata import high_throughput_predict
    
    ad = AgatDatabase(mode_of_NN='ase_dist', num_of_cores=16)
    ad.build()
    
    at = Train()
    at.fit_energy_model()
    at.fit_force_model()
    
    HA = HpAds()
    HA.run(formula, fmax=0.05, steps=200, gpu=0)



# -*- coding: utf-8 -*-
"""
Created on Mon Jun 26 23:07:24 2023

@author: ZHANG Jun
"""

'''

.. py:data:: config

    Configurations for database construction, training process, and prediction behaviors.

'''

import os
import tensorflow as tf
import dgl
dgl.use_libxsmm(False)

default_elements = ['Ac', 'Ag', 'Al', 'Am', 'Ar', 'As', 'At', 'Au', 'B',  'Ba',
                    'Be', 'Bh', 'Bi', 'Bk', 'Br', 'C',  'Ca', 'Cd', 'Ce', 'Cf',
                    'Cl', 'Cm', 'Cn', 'Co', 'Cr', 'Cs', 'Cu', 'Db', 'Ds', 'Dy',
                    'Er', 'Es', 'Eu', 'F',  'Fe', 'Fl', 'Fm', 'Fr', 'Ga', 'Gd',
                    'Ge', 'H',  'He', 'Hf', 'Hg', 'Ho', 'Hs', 'I',  'In', 'Ir',
                    'K',  'Kr', 'La', 'Li', 'Lr', 'Lu', 'Lv', 'Mc', 'Md', 'Mg',
                    'Mn', 'Mo', 'Mt', 'N',  'Na', 'Nb', 'Nd', 'Ne', 'Nh', 'Ni',
                    'No', 'Np', 'O',  'Og', 'Os', 'P',  'Pa', 'Pb', 'Pd', 'Pm',
                    'Po', 'Pr', 'Pt', 'Pu', 'Ra', 'Rb', 'Re', 'Rf', 'Rg', 'Rh',
                    'Rn', 'Ru', 'S',  'Sb', 'Sc', 'Se', 'Sg', 'Si', 'Sm', 'Sn',
                    'Sr', 'Ta', 'Tb', 'Tc', 'Te', 'Th', 'Ti', 'Tl', 'Tm', 'Ts',
                    'U',  'V',  'W',  'Xe', 'Y',  'Yb', 'Zn', 'Zr']

default_data_config =  {
    'species': default_elements,
    'path_file': 'paths.log', # A file of absolute paths where OUTCAR and XDATCAR files exist.
    'build_properties': {'energy': True,
                         'forces': True,
                         'cell': True,
                         'cart_coords': True,
                         'frac_coords': True,
                         'constraints': True,
                         'distance': True,
                         'direction': True,
                         'path': False}, # Properties needed to be built into graph.
    'dataset_path': 'dataset', # Path where the collected data to save.
    'mode_of_NN': 'ase_natural_cutoffs', # How to identify connections between atoms. 'ase_natural_cutoffs', 'pymatgen_dist', 'ase_dist', 'voronoi'. Note that pymatgen is much faster than ase.
    'cutoff': 5.0, # Cutoff distance to identify connections between atoms. Deprecated if ``mode_of_NN`` is ``'ase_natural_cutoffs'``
    'load_from_binary': False, # Read graphs from binary graphs that are constructed before. If this variable is ``True``, these above variables will be depressed.
    'num_of_cores': 2,
    'super_cell': False,
    'has_adsorbate': False,
    'keep_readable_structural_files': False,
    'mask_similar_frames': False,
    'energy_stride': 0.05,
    'scale_prop': False,
    'validation_size': 0.15, # int or float. int: number of samples of the validation set. float: portion of samples of the validation set.
    'test_size': 0.15,
    'new_split': True
             }

default_train_config = {
    'dataset_path': 'dataset',
    'train_energy_model': True,
    'train_force_model': True,
    'epochs': 1000,
    'output_files': 'out_file',
    'new_energy_train': False,
    'new_force_train': False,
    'gpu_for_energy_train': 0,
    'gpu_for_force_train': 0,
    'load_graphs_on_gpu': False, # deprecated
    'validation_size': 0.15,
    'test_size': 0.15,
    'early_stop': True,
    'stop_patience': 300,
    'energy_GAT_out_list': [100, 100, 100],
    'force_GAT_out_list': [100, 100, 100],
    'energy_model_head_list': ['mul', 'div', 'free'],
    'force_model_head_list': ['mul', 'div', 'free'],
    'energy_model_readout_list': [200, 100, 50, 30, 10, 3, 1],
    'force_model_readout_list': [200, 100, 50, 30, 10, 3, 1],
    'bias': True,
    'negative_slope': 0.2,
    'embed_activation': 'LeakyReLU',
    'readout_activation': 'LeakyReLU',
    'energy_loss_fcn': tf.keras.losses.MeanSquaredError(),
    'force_loss_fcn': tf.keras.losses.MeanSquaredError(),
    'energy_optimizer': tf.keras.optimizers.Adam(learning_rate=0.0005, epsilon=1e-8), # Default lr: 0.001
    'force_optimizer': tf.keras.optimizers.Adam(learning_rate=0.0005, epsilon=1e-8), # Default lr: 0.001
    'weight_decay': 5e-5,
    'mae': tf.keras.losses.MeanAbsoluteError(),
    'batch_size': 64,
    'val_batch_size': 400,
    'L2_reg': False,
    'validation_freq': 20000,
    'validation_samples': 20000,
    # 'log_file': 'v7_energy.log',
    'transfer_learning': False,
    'trainable_layers': -4,
    'split_binary_graph': False,
    'batch_normalization': False,
    'mask_fixed': False,
    'tail_readout_noact': 3,
    'adsorbate': False, # indentify adsorbate or not when building graphs.
    'adsorbate_coeff': 20.0 # the importance of adsorbate atoms with respective to surface atoms.
    }

default_ase_calculator_config = {'fmax'             : 0.1,
                                 'steps'            : 200,
                                 'maxstep'          : 0.05,
                                 'restart'          : None,
                                 'restart_steps'    : 0,
                                 'perturb_steps'    : 0,
                                 'perturb_amplitude': 0.05,
                                 'out'              : None }

default_hp_config = {
        'volume_per_atom': 14.045510416666668,
        'energy_model_save_dir': os.path.join('out_file', 'energy_ckpt'),
        'force_model_save_dir': os.path.join('out_file', 'force_ckpt'),
        'opt_config': default_ase_calculator_config,
        'calculation_index'    : '0', # sys.argv[1],
        'fix_all_surface_atom' : False,
        'remove_bottom_atoms'  : False,
        'save_trajectory'      : False,
        'partial_fix_adsorbate': True,
        'adsorbates'           : ['H'],
        'sites'                : ['ontop'],
        'fmax'                 : 0.1,
        'dist_from_surf'       : 1.7,
        'using_template_bulk_structure': False,
        'graph_build_scheme_dir': os.path.join('dataset'),
        'gpu': 0 # in our test results, the A6000 is about 5 times faster than EPYC 7763.
        }

default_predict_config = {}

default_active_learning_config = {}



# -*- coding: utf-8 -*-
"""
Created on Sun Jun 25 22:54:47 2023

@author: ZHANG Jun
"""

import numpy as np
import os
import dgl
import tensorflow as tf
from sklearn.model_selection import train_test_split
from dgl.data.utils import save_graphs, load_graphs
import multiprocessing
from tqdm import tqdm
import json
from ase.io import read, write
import ase
from ase.neighborlist import natural_cutoffs

import sys
from agat.default_parameters import default_data_config
from agat.data.AtomicFeatures import get_atomic_feature_onehot, get_atomic_features
from agat.lib.GatLib import config_parser

class CrystalGraph(object):
    """
    .. py:class:: CrystalGraph(object)

        :param **data_config: Configuration file for building database.
        :type **data_config: str/dict
        :return: A ``DGL.graph``.
        :rtype: ``DGL.graph``.

    .. Hint::
        Although we recommend representing atoms with one hot code, you can
        use the another way with: ``self.all_atom_feat = get_atomic_features()``

    .. Hint::
        In order to build a reasonable graph, a samll cell should be repeated.
        One can modify "self._cell_length_cutoff" for special needs.

    .. Hint::
        We encourage you to use ``ase`` module to build crystal graphs.
        The ``pymatgen`` module needs some dependencies that conflict with
        other modules.

    """
    def __init__(self, **data_config):
        self.data_config = {**default_data_config, **config_parser(data_config)}
        self.all_atom_feat = get_atomic_feature_onehot(self.data_config['species'])

        """ Although we recommend representing atoms with one hot code, you can
        use the another way with the next line: """
        # self.all_atom_feat = get_atomic_features()

        """
        In order to build a reasonable graph, a samll cell should be repeated.
        One can modify "self._cell_length_cutoff" for special needs.
        """
        self._cell_length_cutoff = np.array([11.0, 5.5, 0.0]) # Unit: angstrom. # increase cutoff to repeat more.

        if self.data_config['mode_of_NN'] == 'voronoi' or self.data_config['mode_of_NN'] == 'pymatgen_dist':
            ''' We encourage you to use ``ase`` module to build crystal graphs.
            The ``pymatgen`` module needs some dependencies that conflict with
            other modules.'''
            try:
                from pymatgen.core.structure import Structure
                from agat.data.PymatgenStructureAnalyzer import VoronoiConnectivity
            except ModuleNotFoundError:
                raise ModuleNotFoundError('We encourage you to use ``ase`` module \
                to build crystal graphs. The ``pymatgen`` module needs some \
                dependencies that conflict with other modules. Install\
                ``pymatgen`` with ``pip install pymatgen``.')

            self.Structure = Structure
            self.VoronoiConnectivity = VoronoiConnectivity
        if self.data_config['build_properties']['path']:
            print('You choose to store the path of each graph. \
            Be noted that this will expose your file structure when you \
            publish your dataset.')

    def get_adsorbate_bool(self, element_list):
        """
       .. py:method:: get_adsorbate_bool(self)

          Identify adsorbates based on elements： H and O.

          :return: a list of bool values.
          :rtype: tf.constant

        """
        # element_list = np.array(self.data_config['species'])
        element_list = np.array(element_list)
        return tf.constant(np.where((element_list == 'H') | (element_list == 'O'), True, False), dtype='bool')

    def get_crystal(self, crystal_fpath):
        """
        .. py:method:: get_crystal(self, crystal_fpath)

           Read structural file and return a pymatgen crystal object.

           :param str crystal_fpath: the path to the crystal structural.
           :return: a pymatgen structure object.
           :rtype: ``pymatgen.core.structure``.

        .. Hint:: If there is only one site in the cell, this site has no other
        neighbors. Thus, the graph cannot be built without repetition.

        """
        assert os.path.exists(crystal_fpath), str(crystal_fpath) + " not found."
        structure = self.Structure.from_file(crystal_fpath)
        """If there is only one site in the cell, this site has no other
        neighbors. Thus, the graph cannot be built without repetition."""
        cell_span = structure.lattice.matrix.diagonal()
        if cell_span.min() < max(self._cell_length_cutoff) and self.data_config['super_cell']:
            repeat = np.digitize(cell_span, self._cell_length_cutoff) + 1
            structure.make_supercell(repeat)
        return structure # pymatgen format

    def get_1NN_pairs_voronoi(self, crystal):
        """
        .. py:method:: get_1NN_pairs_voronoi(self, crystal)

           The ``get_connections_new()`` of ``VoronoiConnectivity`` object is modified.

           :param pymatgen.core.structure crystal: a pymatgen structure object.
           :Returns:
              - index of senders
              - index of receivers
              - a list of distance between senders and receiver

        """

        crystal_connect = self.VoronoiConnectivity(crystal, self.data_config['cutoff'])
        return crystal_connect.get_connections_new() # the outputs are bi-directional and self looped.

    def get_1NN_pairs_distance(self, crystal):
        """

        .. py:method:: get_1NN_pairs_distance(self, crystal)

           Find the index of senders, receivers, and distance between them based on the ``distance_matrix`` of pymargen crystal object.

           :param pymargen.core.structure crystal: pymargen crystal object
           :Returns:
              - index of senders
              - index of receivers
              - a list of distance between senders and receivers

        """

        distance_matrix = crystal.distance_matrix
        sender, receiver = np.where(distance_matrix < self.data_config['cutoff'])
        dist = distance_matrix[(sender, receiver)]
        return sender, receiver, dist # the outputs are bi-directional and self looped.

    def get_1NN_pairs_ase_distance(self, ase_atoms):
        """

        .. py:method:: get_1NN_pairs_ase_distance(self, ase_atoms)

           :param ase.atoms ase_atoms: ``ase.atoms`` object.
           :Returns:
              - index of senders
              - index of receivers
              - a list of distance between senders and receivers

        """

        distance_matrix = ase_atoms.get_all_distances(mic=True)
        sender, receiver = np.where(distance_matrix < self.data_config['cutoff'])
        dist = distance_matrix[(sender, receiver)]
        return sender, receiver, dist # the outputs are bi-directional and self looped.

    def get_ndata(self, crystal):
        """

        .. py:method:: get_ndata(self, crystal)

           :param pymargen.core.structure crystal: a pymatgen crystal object.
           :return: ndata: the atomic representations of a crystal graph.
           :rtype: numpy.ndarray

        """

        ndata = []
        if isinstance(crystal, ase.Atoms):
            # num_sites = len(crystal)
            for i in crystal.get_chemical_symbols():
                ndata.append(self.all_atom_feat[i])
        else: # isinstance(crystal, self.Structure)
            num_sites = crystal.num_sites
            for i in range(num_sites):
                ndata.append(self.all_atom_feat[crystal[i].species_string])

        # else:
        #     print(f'Wraning!!! Unrecognized structure type: {crystal}')

        return np.array(ndata)

    def get_graph_from_ase(self, fname): # ase_atoms or file name
        '''
        .. py:method:: get_graph_from_ase(self, fname)

           Build graphs with ``ase``.

           :param str/``ase.Atoms`` fname: File name or ``ase.Atoms`` object.
           :return: A bidirectional graph with self-loop connection.
           :return: A dict of information of graph-level features.
        '''

        assert not isinstance(fname, ase.Atoms) or not self.data_config['build_properties']['forces'], 'The input cannot be a ase.Atoms object when include_forces is ``True``, try with an ``OUTCAR``'

        if not isinstance(fname, ase.Atoms):
            ase_atoms = read(fname)
        else:
            ase_atoms = fname
        num_sites = len(ase_atoms)
        if self.data_config['mode_of_NN'] == 'ase_dist':
            assert self.data_config['cutoff'], 'The `cutoff` cannot be `None` \
            in this case. Provide a float here.'
            ase_cutoffs = self.data_config['cutoff']
        elif self.data_config['mode_of_NN'] == 'ase_natural_cutoffs':
            ase_cutoffs = np.array(natural_cutoffs(ase_atoms, mult=1.25, H=3.0, O=3.0)) # the specified values are radius, not diameters.
        i, j, d, D = ase.neighborlist.neighbor_list('ijdD', # i: sender; j: receiver; d: distance; D: direction
                                                    ase_atoms,
                                                    cutoff=ase_cutoffs,
                                                    self_interaction=True)

        ndata                   = self.get_ndata(ase_atoms)
        bg                      = dgl.graph((i, j))
        bg.ndata['h']           = tf.constant(ndata, dtype='float32')
        if self.data_config['build_properties']['distance']:
            bg.edata['dist']        = tf.constant(d, dtype='float32')
        if self.data_config['build_properties']['direction']:
            bg.edata['direction']   = tf.constant(D, dtype='float32')
        if self.data_config['build_properties']['constraints']:
            constraints             = [[True, True, True]] * num_sites
            for c in ase_atoms.constraints:
                if isinstance(c, ase.constraints.FixScaled):
                    constraints[c.a] = c.mask
                elif isinstance(c, ase.constraints.FixAtoms):
                    for i in c.index:
                        constraints[i] = [False, False, False]
                elif isinstance(c, ase.constraints.FixBondLengths):
                    pass
                else:
                    raise TypeError(f'Wraning!!! Undefined constraint type: {type(c)}')
            bg.ndata['constraints'] = tf.constant(constraints, dtype='bool')
        if self.data_config['build_properties']['forces']:
            forces_true             = tf.constant(np.load(fname+'_force.npy'), dtype='float32')
            bg.ndata['forces_true'] = tf.constant((forces_true), dtype='float32')
        if self.data_config['build_properties']['cart_coords']:
            bg.ndata['cart_coords'] = tf.constant(ase_atoms.positions, dtype='float32')
        if self.data_config['build_properties']['frac_coords']:
            bg.ndata['frac_coords'] = tf.constant(ase_atoms.get_scaled_positions(), dtype='float32')
        if self.data_config['has_adsorbate']:
            element_list            = ase_atoms.get_chemical_symbols()
            bg.ndata['adsorbate']   = self.get_adsorbate_bool(element_list)

        graph_info = {}
        if self.data_config['build_properties']['energy']:
            energy_true = tf.constant(np.load(fname+'_energy.npy'), dtype='float32')
            graph_info['energy_true'] = tf.constant((energy_true), dtype='float32')
        if self.data_config['build_properties']['cell']:
            cell_true = tf.constant(ase_atoms.cell.array, dtype='float32')
            graph_info['cell_true'] = tf.constant((cell_true), dtype='float32')
        if self.data_config['build_properties']['path']:
            graph_info['path'] = fname

        return bg, graph_info

    def get_graph_from_pymatgen(self, crystal_fname):
        """
        .. py:method:: get_graph_from_pymatgen(self, crystal_fname)

           Build graphs with pymatgen.

           :param str crystal_fname: File name.
           :return: A bidirectional graph with self-loop connection.
           :return: A dict of information of graph-level features.
        """
        assert bool(not bool(self.data_config['super_cell'] and self.data_config['build_properties']['forces'])), 'include_forces cannot be True when super_cell is True.'
        # https://docs.dgl.ai/guide_cn/graph-feature.html 通过张量分配创建特征时，DGL会将特征赋给图中的每个节点和每条边。该张量的第一维必须与图中节点或边的数量一致。 不能将特征赋给图中节点或边的子集。
        mycrystal               = self.get_crystal(crystal_fname)
        # self.cart_coords       = mycrystal.cart_coords           # reserved for other functions

        if self.data_config['mode_of_NN'] == 'voronoi':
            sender, receiver, dist  = self.get_1NN_pairs_voronoi(mycrystal)   # the dist array is bidirectional and self-looped. # dist需要手动排除cutoff外的邻居
            location                = np.where(np.array(dist) > self.data_config['cutoff'])
            sender                  = np.delete(np.array(sender), location)
            receiver                = np.delete(np.array(receiver), location)
            dist                    = np.delete(np.array(dist), location)
        elif self.data_config['mode_of_NN'] == 'pymatgen_dist':
            sender, receiver, dist  = self.get_1NN_pairs_distance(mycrystal)

        ndata                   = self.get_ndata(mycrystal)
        bg                      = dgl.graph((sender, receiver))
        bg.ndata['h']           = tf.constant(ndata, dtype='float32')

        if self.data_config['build_properties']['distance']:
            bg.edata['dist']        = tf.constant(dist, dtype='float32')
        if self.data_config['build_properties']['cart_coords']:
            bg.ndata['cart_coords'] = tf.constant(mycrystal.cart_coords, dtype='float32')
        if self.data_config['build_properties']['frac_coords']:
            bg.ndata['frac_coords'] = tf.constant(mycrystal.frac_coords, dtype='float32')
        if self.data_config['build_properties']['direction']:
            frac_coords             = mycrystal.frac_coords
            sender_frac_coords      = frac_coords[sender]
            receiver_frac_coords    = frac_coords[receiver]
            delta_frac_coords       = receiver_frac_coords - sender_frac_coords
            image_top               = np.where(delta_frac_coords < -0.5, -1.0, 0.0)
            image_bottom            = np.where(delta_frac_coords >  0.5,  1.0, 0.0)
            image                   = image_top + image_bottom
            sender_frac_coords     += image
            sender_cart_coords      = mycrystal.lattice.get_cartesian_coords(sender_frac_coords)
            receiver_cart_coords    = mycrystal.cart_coords[receiver]
            real_direction          = receiver_cart_coords - sender_cart_coords
            direction_norm          = np.linalg.norm(real_direction, axis=1, keepdims=True)
            direction_normalized    = real_direction / direction_norm
            direction_normalized[np.where(sender == receiver)] = 0.0
            bg.edata['direction']   = tf.constant(direction_normalized, dtype='float32')
        if self.data_config['build_properties']['constraints']:
            try:
                bg.ndata['constraints'] = tf.constant([x.selective_dynamics for x in mycrystal], dtype='bool')
            except AttributeError:
                bg.ndata['constraints'] = tf.constant([[True, True, True] for x in mycrystal], dtype='bool')
        if self.data_config['build_properties']['forces']:
            forces_true             = tf.constant(np.load(crystal_fname+'_force.npy'), dtype='float32')
            bg.ndata['forces_true'] = tf.constant((forces_true), dtype='float32')
        if self.data_config['has_adsorbate']:
            element_list            = [x.specie.name for x in mycrystal.sites]
            bg.ndata['adsorbate']   = self.get_adsorbate_bool(element_list)

        graph_info = {}
        if self.data_config['build_properties']['energy']:
            energy_true = tf.constant(np.load(crystal_fname+'_energy.npy'), dtype='float32')
            graph_info['energy_true'] = tf.constant((energy_true), dtype='float32')
        if self.data_config['build_properties']['cell']:
            cell_true = tf.constant(mycrystal.lattice.matrix, dtype='float32')
            graph_info['cell_true'] = tf.constant((cell_true), dtype='float32')
        if self.data_config['build_properties']['path']:
            graph_info['path'] = crystal_fname
        return bg, graph_info

    def get_graph(self, crystal_fname):
        """
        .. py:method:: get_graph(self, crystal_fname)

           This method can choose which graph-construction method is used, according to the ``mode_of_NN`` attribute.

           .. Hint:: You can call this method to build one graph.

           :param str crystal_fname: File name.
           :return: A bidirectional graph with self-loop connection.
        """
        if self.data_config['mode_of_NN'] == 'voronoi' or self.data_config['mode_of_NN'] == 'pymatgen_dist':
            return self.get_graph_from_pymatgen(crystal_fname)
        elif self.data_config['mode_of_NN'] == 'ase_natural_cutoffs' or self.data_config['mode_of_NN'] == 'ase_dist':
            return self.get_graph_from_ase(crystal_fname)

class ReadGraphs(object):
    """
    .. py:class:: CrystalGraph(object)

        :param **data_config: Configuration file for building database.
        :type **data_config: str/dict
        :return: A ``DGL.graph``.
        :rtype: ``DGL.graph``.

    """
    def __init__(self, **data_config):
        self.data_config = {**default_data_config, **config_parser(data_config)}
        assert os.path.exists(self.data_config['dataset_path']), str(self.data_config['dataset_path']) + " not found."

        self.cg               = CrystalGraph(**data_config)
        fname_prop_data       = np.loadtxt(os.path.join(self.data_config['dataset_path'],
                                                        'fname_prop.csv'),
                                           dtype=str, delimiter=',')
        self.name_list        = fname_prop_data[:,0]
        self.number_of_graphs = len(self.name_list)

    def read_batch_graphs(self, batch_index_list, batch_num):
        batch_fname  = [self.name_list[x] for x in batch_index_list]
        batch_g, batch_graph_info = [], []
        for fname in tqdm(batch_fname, desc='Reading ' + str(batch_num) + ' batch graphs', delay=batch_num):
            g, graph_info = self.cg.get_graph(os.path.join(self.data_config['dataset_path'] ,fname))
            batch_g.append(g)
            batch_graph_info.append(graph_info)

        batch_labels = {key: tf.stack([i[key] for i in batch_graph_info])\
                        for key in batch_graph_info[0] if key != 'path'}
        # print(batch_labels)
        save_graphs(os.path.join(self.data_config['dataset_path'], 'all_graphs_' + str(batch_num) + '.bin'), batch_g, batch_labels)

    def read_all_graphs(self): # prop_per_node=False Deprecated!
        """
        .. py:method:: read_all_graphs(self)

           Read all graphs specified in the csv file.

           .. Note:: The loaded graphs are saved under the attribute of :py:attr:`dataset_path`.

           .. DANGER:: Do not scale the label if you don't know what are you doing.

           :param bool scale_prop: scale the label or not. DO NOT scale unless you know what you are doing.
           :param str ckpt_path: checkpoint directory of the well-trained model.
           :Returns:
              - graph_list： a list of ``DGL`` graph.
              - graph_labels： a list of labels.
        """
        if self.data_config['load_from_binary']:
            try:
                graph_path = os.readlink(os.path.join(self.data_config['dataset_path'], 'all_graphs.bin'))
            except:
                graph_path = 'all_graphs.bin'
            cwd = os.getcwd()
            os.chdir(self.data_config['dataset_path'])
            graph_list, graph_labels = load_graphs(graph_path)
            os.chdir(cwd)
        else:
            num_graph_per_core = self.number_of_graphs // self.data_config['num_of_cores'] + 1
            graph_index        = [x for x in range(self.number_of_graphs)]
            batch_index        = [graph_index[x: x + num_graph_per_core] for x in range(0, self.number_of_graphs, num_graph_per_core)]
            p                  = multiprocessing.Pool(self.data_config['num_of_cores'])

            for batch_num, batch_index_list in enumerate(batch_index):
                p_out = p.apply_async(self.read_batch_graphs, args=(batch_index_list, batch_num))
            p_out.get()
            print('Waiting for all subprocesses...')
            p.close()
            p.join()
            print('All subprocesses done.')
            graph_list = []
            graph_labels = {}
            for x in range(self.data_config['num_of_cores']):
                batch_g, batch_labels = load_graphs(os.path.join(self.data_config['dataset_path'],
                                                                 'all_graphs_' + str(x) + '.bin'))
                graph_list.extend(batch_g)
                for key in batch_labels.keys():
                    try:
                        # graph_labels[key].extend(batch_labels[key])
                        graph_labels[key] = tf.concat([graph_labels[key],
                                                       batch_labels[key]], 0)
                    except KeyError:
                        graph_labels[key] = batch_labels[key]

                os.remove(os.path.join(self.data_config['dataset_path'], 'all_graphs_' + str(x) + '.bin'))
            save_graphs(os.path.join(self.data_config['dataset_path'], 'all_graphs.bin'), graph_list, graph_labels)

            with open(os.path.join(self.data_config['dataset_path'], 'graph_build_scheme.json'), 'w') as fjson:
                json.dump(self.data_config, fjson, indent=4)
        return graph_list, graph_labels

class TrainValTestSplit(object):
    """
    Description:
    ----------
        Split the dataset.
    Parameters
    ----------
    validation_size: int or float
        int: number of samples of the validation set.
        float: portion of samples of the validation set
    test_size: int or float
        int: number of samples of the validation set.
        float: portion of samples of the validation set
    csv_file: str
        File name of a csv file that contains the filenames of crystals
        with cif or VASP formate.
    new_split: boolean
        Split the dataset by `sklearn.model_selection.train_test_split` or
        loaded from previously saved txt files.
    Returns of `__call__` method
    ----------------------------
    train_index : list
        A list of integers of training dataset.
    validation_index : list
        A list of integers of validation dataset.
    test_index : list
        A list of integers of test dataset.
    """
    def __init__(self, **data_config):
        self.data_config = {**default_data_config, **config_parser(data_config)}

        fname_prop_data       = np.loadtxt(os.path.join(self.data_config['dataset_path'],
                                                        'fname_prop.csv'),
                                           dtype=str, delimiter=',')
        self.number_of_graphs = np.shape(fname_prop_data)[0]

    def __call__(self):
        if self.data_config['new_split']:
            train_index,      validation_and_test_index = train_test_split([x for x in range(self.number_of_graphs)],
                                                                           test_size=self.data_config['test_size']+self.data_config['validation_size'],
                                                                           shuffle=True)
            validation_index, test_index                = train_test_split(validation_and_test_index,
                                                                           test_size=self.data_config['test_size']/(self.data_config['test_size']+self.data_config['validation_size']),
                                                                           shuffle=True)
            np.savetxt(os.path.join(self.data_config['dataset_path'], 'train.txt'),      train_index,      fmt='%.0f')
            np.savetxt(os.path.join(self.data_config['dataset_path'], 'validation.txt'), validation_index, fmt='%.0f')
            np.savetxt(os.path.join(self.data_config['dataset_path'], 'test.txt'),       test_index,       fmt='%.0f')
        else:
            try:
                train_index      = np.loadtxt(os.path.join(self.data_config['dataset_path'], 'train.txt'),      dtype=int)
                validation_index = np.loadtxt(os.path.join(self.data_config['dataset_path'], 'validation.txt'), dtype=int)
                test_index       = np.loadtxt(os.path.join(self.data_config['dataset_path'], 'test.txt'),       dtype=int)
            except OSError:
                print('User: Index file not found, generate new files...')
                train_index,      validation_and_test_index = train_test_split([x for x in range(self.number_of_graphs)],
                                                                               test_size=self.data_config['test_size']+self.data_config['validation_size'],
                                                                               shuffle=True)
                validation_index, test_index                = train_test_split(validation_and_test_index,
                                                                               test_size=self.data_config['test_size']/(self.data_config['test_size']+self.data_config['validation_size']),
                                                                               shuffle=True)
                np.savetxt(os.path.join(self.data_config['dataset_path'], 'train.txt'),      train_index,      fmt='%.0f')
                np.savetxt(os.path.join(self.data_config['dataset_path'], 'validation.txt'), validation_index, fmt='%.0f')
                np.savetxt(os.path.join(self.data_config['dataset_path'], 'test.txt'),       test_index,       fmt='%.0f')
        return train_index, validation_index, test_index

class ExtractVaspFiles(object):
    '''
    :param data_config['dataset_path']: Absolute path where the collected data to save.
    :type data_config['dataset_path']: str

    .. Note:: Always save the property per node as the label. For example: energy per atom (eV/atom).

    '''
    def __init__(self, **data_config):
        self.data_config = {**default_data_config, **config_parser(data_config)}
        if not os.path.exists(self.data_config['dataset_path']):
            os.mkdir(self.data_config['dataset_path'])

        self.in_path_list = np.loadtxt(self.data_config['path_file'], dtype=str)

        self.working_dir = os.getcwd()

    def read_oszicar(self,fname='OSZICAR'):
        ee_steps = []
        with open(fname, 'r') as f:
            lines = f.readlines()

        for i, line in enumerate(lines):
            if 'E0=' in line.split():
                ee_steps.append(int(lines[i-1].split()[1]))
        return ee_steps

    def read_incar(self, fname='INCAR'):
        with open(fname, 'r') as f:
            lines = f.readlines()
        for line in lines:
            if 'NELM' in line.split():
                NELM = int(line.split()[2])
                break
        return NELM

    def split_output(self, process_index):
        '''
        :param process_index: A number to index the process.
        :type process_index: int.

        '''

        print('mask_similar_frames:', self.data_config['mask_similar_frames'])
        f_csv = open(os.path.join(self.data_config['dataset_path'], f'fname_prop_{process_index}.csv'), 'w', buffering=1)
        for path_index, in_path in tqdm(enumerate(self.in_path_list), desc='Extracting ' + str(process_index) + ' VASP files.', delay=process_index): # tqdm(batch_fname, desc='Reading ' + str(batch_num) + ' batch graphs', delay=batch_num)
            in_path = in_path.strip("\n")

            os.chdir(in_path)

            if os.path.exists('OUTCAR') and os.path.exists('XDATCAR') and os.path.exists('OSZICAR') and os.path.exists('INCAR') and os.path.exists('CONTCAR'):

                # read free energy
                with open('OUTCAR', 'r') as f:
                    free_energy = []
                    for line in f:
                        if '  free  energy   TOTEN  ' in line.split('='):
                            free_energy.append(float(line.split()[4]))

                read_good = True
                try:
                    # read frames
                    frame_contcar  = read('CONTCAR')
                    constraints    = frame_contcar.constraints
                    frames_outcar  = read('OUTCAR', index=':')   # coordinates in OUTCAR file are less accurate than that in XDATCAR. Energy in OUTCAR file is more accurate than that in OSZICAR file
                    frames_xdatcar = read('XDATCAR', index=':')
                    [x.set_constraint(constraints) for x in frames_xdatcar]

                    num_atoms      = len(frames_outcar[0])
                    num_frames     = len(frames_outcar)
                    assert len(frames_outcar) == len(frames_xdatcar), f'Inconsistent number of frames between OUTCAR and XDATCAR files. OUTCAR: {len(frames_outcar)}; XDATCAR {len(frames_xdatcar)}'
                    assert num_frames == len(free_energy), 'Number of frams does not equal to number of free energies.'
                    ee_steps = self.read_oszicar()
                    assert len(ee_steps) == len(frames_xdatcar), f'Inconsistent number of frames between OSZICAR and XDATCAR files. OSZICAR: {len(ee_steps)}; XDATCAR {len(frames_xdatcar)}'
                    NELM     = self.read_incar()
                except:
                    print(f'Read OUTCAR, OSZICAR, INCAR, CONTCAR, and/or XDATCAR with exception in: {in_path}')
                    read_good = False

                if read_good:
                    ###############################################################
                    # mask similar frames based on energy difference.
                    if self.data_config['mask_similar_frames']:
                        no_mask_list = [0]
                        for i, e in enumerate(free_energy):
                            if abs(e - free_energy[no_mask_list[-1]]) > self.data_config['energy_stride']:
                                no_mask_list.append(i)
                        if not no_mask_list[-1] == num_frames - 1:
                            no_mask_list.append(num_frames - 1)

                        free_energy    = [free_energy[x] for x in no_mask_list]
                        frames_outcar  = [frames_outcar[x] for x in no_mask_list]
                        frames_xdatcar = [frames_xdatcar[x] for x in no_mask_list]
                        ee_steps       = [ee_steps[x] for x in no_mask_list]
                    else:
                        no_mask_list = [x for x in range(num_frames)]
                    ###############################################################

                    free_energy_per_atom = [x / num_atoms for x in free_energy]

                    # save frames
                    for i in range(len(no_mask_list)):
                        if ee_steps[i] < NELM:
                            fname = str(os.path.join(self.data_config['dataset_path'], f'POSCAR_{process_index}_{path_index}_{i}'))
                            while os.path.exists(os.path.join(self.working_dir, fname)):
                                fname = fname + '_new'

                            frames_xdatcar[i].write(os.path.join(self.working_dir, fname))
                            forces = frames_outcar[i].get_forces(apply_constraint=False)
                            np.save(os.path.join(self.working_dir, f'{fname}_force.npy'), forces)
                            np.save(os.path.join(self.working_dir,f'{fname}_energy.npy'), free_energy_per_atom[i])
                            f_csv.write(os.path.basename(fname) + ',  ')
                            f_csv.write(str(free_energy_per_atom[i]) + ',  ' + str(in_path) + '\n')
                        else:
                            print(f'Electronic steps of {i}th ionic step greater than NELM: {NELM} in path: {in_path}')

            else:
                print(f'OUTCAR, OSZICAR, INCAR, CONTCAR, and/or XDATCAR files do not exist in {in_path}.')

            os.chdir(self.working_dir)
        f_csv.close()

    def __call__(self):
        batch_index = np.array_split([x for x in range(len(self.in_path_list))], self.data_config['num_of_cores'])

        processes = []
        for process_index, path_index in enumerate(batch_index):
            path_list = [self.in_path_list[x] for x in path_index]
            p = multiprocessing.Process(target=self.split_output, args=[process_index])
            p.start()
            processes.append(p)
        print(processes)

        for process in processes:
            process.join()

        f = open(os.path.join(self.working_dir,
                              self.data_config['dataset_path'],
                              'fname_prop.csv'), 'w')
        for job, _ in enumerate(batch_index):
            lines = np.loadtxt(os.path.join(self.working_dir,
                                            self.data_config['dataset_path'],
                                            f'fname_prop_{job}.csv'),
                               dtype=str)
            np.savetxt(f, lines, fmt='%s')
        f.close()

class AgatDatabase():
    def __init__(self, **data_config):
        self.data_config = {**default_data_config, **config_parser(data_config)}

    def build(self):
        # extract vasp files.
        evf = ExtractVaspFiles(**self.data_config)()

        # build binary DGL graphs.
        graph_reader = ReadGraphs(**self.data_config)
        graph_reader.read_all_graphs()

        # split the dataset.
        train_index, validation_index, test_index = TrainValTestSplit(**self.data_config)()
        # print(os.getcwd())
        # curdir = os.getcwd()
        if not self.data_config['keep_readable_structural_files']:
            fname_prop_data = np.loadtxt(os.path.join(self.data_config['dataset_path'],
                                                      'fname_prop.csv'),
                                         dtype=str, delimiter=',')
            fname_list      = fname_prop_data[:,0]
            for fname in fname_list:
                os.remove(os.path.join(self.data_config['dataset_path'], fname))
                os.remove(os.path.join(self.data_config['dataset_path'], f'{fname}_energy.npy'))
                os.remove(os.path.join(self.data_config['dataset_path'], f'{fname}_force.npy'))
            # os.remove(os.path.join(self.data_config['dataset_path'], 'fname_prop.csv'))
            for i in range(self.data_config['num_of_cores']):
                os.remove(os.path.join(self.data_config['dataset_path'], f'fname_prop_{i}.csv'))

def concat_graphs(*list_of_bin):
    """ Concat binary graph files.

    :param *list_of_bin: input file names of binary graphs.
    :type *list_of_bin: strings
    :return: A new file is saved to the current directory: concated_graphs.bin.
    :rtype: None. A new file.

    Example::

        concat_graphs('graphs1.bin', 'graphs2.bin', 'graphs3.bin')

    """

    graph_list = []
    graph_labels = {}
    for file in list_of_bin:
        batch_g, batch_labels = load_graphs(file)
        graph_list.extend(batch_g)
        for key in batch_labels.keys():
            try:
                graph_labels[key] = tf.concat([graph_labels[key],
                                               batch_labels[key]], 0)
            except KeyError:
                graph_labels[key] = batch_labels[key]

    save_graphs('concated_graphs.bin', graph_list, graph_labels)

# build data
if __name__ == '__main__':
    ad = AgatDatabase(mode_of_NN='pymatgen_dist', num_of_cores=16)
    ad.build()




# -*- coding: utf-8 -*-
"""
Created on Tue Dec  7 21:59:30 2021

@author: ZHANG Jun
"""

from ase.optimize import BFGS
from ase.io import read, write
from ase.lattice.cubic import FaceCenteredCubic
from ase.visualize import view
from ase.data import covalent_radii, atomic_numbers # index of `ase.data` starts from ONE.
# from pymatgen.core.ion import Ion
import numpy as np
import os
from ase.build import add_vacuum, sort
from ase.constraints import FixAtoms, FixBondLength
from ase.formula import Formula
import json
import sys

from agat.app.GatApp import GatAseCalculator
from agat.app.cata.generate_adsorption_sites import AddAtoms
from agat.lib.AgatException import file_exit
from agat.lib.file import generate_file_name
from agat.lib.GatLib import config_parser
from agat.default_parameters import default_hp_config
from agat.lib.adsorbate_poscar import adsorbate_poscar

class HpAds(object):
    def __init__(self, **hp_config):
        self.hp_config = {**default_hp_config, **config_parser(hp_config)}

    def perturb_positions(self, atoms, amplitude=0.1):
        calculator  = atoms.get_calculator()
        atoms       = atoms.copy()
        atoms.set_calculator(calculator)
        posistions  = atoms.arrays['positions']
        num_atoms   = len(atoms)
        increment   = np.clip(np.random.normal(0.0, amplitude / 3,  size=(num_atoms,3)), -amplitude, amplitude) # np.random.uniform(-amplitude, amplitude, (num_atoms,3))
        constraints = atoms.constraints
        if len(constraints) > 0:
            increment[constraints[0].index] = 0.0
        posistions += increment
        atoms.set_positions(posistions)
        return atoms

    def scale_atoms(self, atoms, scale_factor=1.0):
        calculator = atoms.get_calculator()
        new_atoms  = atoms.copy()
        new_atoms.set_calculator(calculator)
        cell        = new_atoms.get_cell()
        frac_coords = new_atoms.get_scaled_positions()
        new_cell    = cell * scale_factor
        new_atoms.set_cell(new_cell)
        new_atoms.set_scaled_positions(frac_coords)
        return new_atoms

    def get_concentration_from_ase_formula(self, formula):
        f_dict = Formula(formula).count()
        tot = np.sum(list(f_dict.values()))
        c_dict = {k: v/tot for k, v in f_dict.items()}
        c_dict = {k:c_dict[k] for k in c_dict if c_dict[k] > 0}
        return c_dict

    def get_v_per_atom(self, chemical_formula):
        # for now, use this for NiCoFePdPt only.
        # v_per_atom = -282.7957531391954 * c_NiCoFe(sum) - 278.79605077419797 * C_Pd - -278.6228860885035 * C_Pt + 293.66128761358624
        # comp             = Ion.from_formula(chemical_formula)
        # elements         = [x.name for x in comp.elements]
        # atomic_fracions  = [comp.get_atomic_fraction(x) for x in elements]
        # frac_dict        = dict(zip(elements, atomic_fracions))
        # frac_dict        = {**frac_dict, **{k:0 for k in ['Ni', 'Co', 'Fe', 'Pd', 'Pt']\
        #                                     if not frac_dict.__contains__(k)}}
        frac_dict = self.get_concentration_from_ase_formula(chemical_formula)
        return -282.7957531391954 * (frac_dict['Ni'] + frac_dict['Co'] + frac_dict['Fe'])\
            - 278.79605077419797 * frac_dict['Pd'] - 278.6228860885035 * frac_dict['Pt']\
                + 293.66128761358624

    def get_ase_atom_from_formula(self, chemical_formula, v_per_atom=None):
        # interpret formula
        atomic_fracions = self.get_concentration_from_ase_formula(chemical_formula)
        elements = [x for x in atomic_fracions]
        element_number = [atomic_numbers[x] for x in elements]
        mean_radii       = np.sum([covalent_radii[n] * atomic_fracions[e] for n, e in zip(element_number, elements)]) #

        Pt_radii           = covalent_radii[78]
        latticeconstant    = mean_radii / Pt_radii * 3.92
        atoms              = FaceCenteredCubic('Pt', directions=[[1,-1,0], [1,1,-2], [1,1,1]], size=(4, 3, 2), latticeconstant=latticeconstant, pbc=True)
        total_atom         = len(atoms)
        num_atom_list      = np.array(list(atomic_fracions.values())) * total_atom
        num_atom_list      = np.around(num_atom_list, decimals=0)
        total_tmp          = np.sum(num_atom_list)
        deviation          = total_atom - total_tmp
        num_atom_list[np.random.randint(len(elements))] += deviation

        # shuffle atoms
        ase_number    = []
        for i_index, i in enumerate(num_atom_list):
            for j in range(int(i)):
                ase_number.append(element_number[i_index])
        np.random.shuffle(ase_number)
        atoms.set_atomic_numbers(ase_number)

        # scale atoms
        if isinstance(v_per_atom, (float, int)):
            volume = atoms.cell.volume
            volume_per_atom = volume / len(atoms)
            volume_ratio = v_per_atom / volume_per_atom
            scale_factor = pow(volume_ratio, 1/3)
            atoms = self.scale_atoms(atoms, scale_factor)
        return atoms

    def get_ase_atom_from_formula_template(self, chemical_formula, v_per_atom=None,
                                           template_file='POSCAR_temp'):
        # interpret formula
        # the template file should be a bulk structure
        atomic_fracions    = self.get_concentration_from_ase_formula(chemical_formula)
        elements           = [x for x in atomic_fracions]
        element_number     = [atomic_numbers[x] for x in elements]
        atoms              = read(template_file)
        total_atom         = len(atoms)
        num_atom_list      = np.array(list(atomic_fracions.values())) * total_atom
        num_atom_list      = np.around(num_atom_list, decimals=0)
        total_tmp          = np.sum(num_atom_list)
        deviation          = total_atom - total_tmp
        num_atom_list[np.random.randint(len(elements))] += deviation

        # shuffle atoms
        ase_number    = []
        for i_index, i in enumerate(num_atom_list):
            for j in range(int(i)):
                ase_number.append(element_number[i_index])
        np.random.shuffle(ase_number)
        atoms.set_atomic_numbers(ase_number)

        # scale atoms
        if isinstance(v_per_atom, (float, int)):
            volume = atoms.cell.volume
            volume_per_atom = volume / len(atoms)
            volume_ratio = v_per_atom / volume_per_atom
            scale_factor = pow(volume_ratio, 1/3)
            atoms = self.scale_atoms(atoms, scale_factor)
        return atoms

    def geo_opt(self, atoms_with_calculator, **kwargs):
        calculator = atoms_with_calculator.get_calculator()
        atoms  = atoms_with_calculator.copy()
        atoms.set_calculator(calculator)

        config = {**self.hp_config['opt_config'], **kwargs}

        if isinstance(config["out"], type(None)):
            logfile    = '-'
            trajectory = None
        else:
            logfile    = f'{config["out"]}.log'
            trajectory = f'{config["out"]}.traj'
        force_opt, energy_opt, atoms_list = [], [], []
        for i in range(config["perturb_steps"]+1):
            dyn = BFGS(atoms,
                       logfile=logfile,
                       trajectory=trajectory,
                       restart=config["restart"],
                       maxstep=config["maxstep"])

            return_code  = dyn.run(fmax=config["fmax"], steps=config["steps"])
            restart_step = 0
            while not return_code and restart_step < config["restart_steps"]:
                restart_step += 1
                config["maxstep"]      /= 2.0
                dyn = BFGS(atoms,
                           logfile=logfile,
                           trajectory=trajectory,
                           restart=config["restart"],
                           maxstep=config["maxstep"])
                return_code  = dyn.run(fmax=config["fmax"], steps=config["steps"])
            force_opt.append(atoms.get_forces())
            energy_opt.append(atoms.get_potential_energy(apply_constraint=False))
            atoms_list.append(atoms.copy())
            if config["perturb_steps"] > 0:
                atoms = self.perturb_positions(atoms, amplitude=config["perturb_amplitude"])
        argmin = np.argmin(energy_opt)
        energy, force, atoms = energy_opt[argmin], force_opt[argmin], atoms_list[argmin]
        force_max = np.linalg.norm(force, axis=1).max()
        return energy, force, atoms, force_max

    def ads_calc(self, formula, calculator, **kwargs):
        hp_config =  {**self.hp_config, **kwargs}

        if hp_config['save_trajectory']:
            out_dir = f'{hp_config["calculation_index"]}_th_calculation'
            if not os.path.exists(out_dir):
                os.mkdir(out_dir)
        else:
            out_dir     = '.'
            outbasename = None

        # generate bulk structure
        chemical_formula = formula

        if self.hp_config['using_template_bulk_structure']:
            atoms = self.get_ase_atom_from_formula_template(chemical_formula,
                                                            self.get_v_per_atom(chemical_formula),
                                                            template_file='POSCAR_temp')
        else:
            atoms = self.get_ase_atom_from_formula(chemical_formula)

        atoms.set_calculator(calculator)

        if hp_config['save_trajectory']:
            write(os.path.join(out_dir, 'POSCAR_bulk.gat'), atoms, format='vasp')
            outbasename = os.path.join(out_dir, 'bulk_opt')

        hp_config['out'] = outbasename

        energy_bulk, force_bulk, atoms_bulk, force_max_bulk = self.geo_opt(atoms, **hp_config)
        if hp_config['save_trajectory']:
            write(os.path.join(out_dir, 'CONTCAR_bulk.gat'), atoms_bulk)

        print('Bulk optimization done.')

        # add vacuum space and fix bottom atoms
        len_z = atoms_bulk.cell.array[2][2]
        c     = FixAtoms(indices=np.where(atoms_bulk.positions[:,2] < len_z / 2 - 1.0)[0])
        atoms_bulk.set_constraint(c)

        if hp_config['remove_bottom_atoms']:
            pop_list = np.where(atoms_bulk.positions[:,2] < 1.0)
            del atoms_bulk[pop_list]

        atoms_bulk.positions += 1.3 # avoid PBC error
        atoms_bulk.wrap()
        add_vacuum(atoms_bulk, 10.0)

        if hp_config['save_trajectory']:
            write(os.path.join(out_dir, 'POSCAR_surface.gat'), atoms_bulk)
            outbasename = os.path.join(out_dir, 'surface_opt')

        # surface optimization
        atoms_bulk.set_calculator(calculator)
        hp_config['out'] = outbasename

        energy_surf, force_surf, atoms_surf, force_max_surf = self.geo_opt(atoms_bulk, **hp_config)
        if hp_config['save_trajectory']:
            write(os.path.join(out_dir, 'CONTCAR_surface.gat'), atoms_surf)

        print('Surface optimization done.')

        if True: # force_max_surf < hp_config['fmax']:
            if hp_config['fix_all_surface_atom']:
                c = FixAtoms(indices=[x for x in range(len(atoms_surf))])
                atoms_surf.set_constraint(c)
            write(f'POSCAR_surf_opt_{hp_config["calculation_index"]}.gat', sort(atoms_surf))

            # adsorbate_shift = {'bridge': 0.0, 'ontop': 0.35, 'hollow': -0.1}

            for ads in hp_config['adsorbates']:
                # generate adsorption configurations: OH adsorption
                adder     = AddAtoms(f'POSCAR_surf_opt_{hp_config["calculation_index"]}.gat',
                                     species=ads,
                                     sites=hp_config['sites'],
                                     dist_from_surf=hp_config['dist_from_surf'],
                                     num_atomic_layer_along_Z=6)
                all_sites = adder.write_file_with_adsorption_sites(adsorbate_poscar,
                                                                   calculation_index=hp_config['calculation_index'])

                # adsorption optimization
                ase_atoms = [read(f'POSCAR_{hp_config["calculation_index"]}_{x}') for x in range(all_sites)]
                [os.remove(f'POSCAR_{hp_config["calculation_index"]}_{x}') for x in range(all_sites)]
                # ase_atoms = ase_atoms[0:3] # !!!

                energy_ads_list, converge_stat = [], []
                for i, ads_atoms in enumerate(ase_atoms):
                    file_exit()
                    if hp_config['save_trajectory']:
                        write(os.path.join(out_dir, f'POSCAR_{ads}_ads_{hp_config["calculation_index"]}_{i}.gat'),
                              ads_atoms)
                        outbasename = os.path.join(out_dir, f'adsorption_{ads}_opt_{i}')

                    hp_config['out'] = outbasename
                    ads_atoms.set_calculator(calculator)
                    energy_ads, force_ads, atoms_ads, force_max_ads = self.geo_opt(ads_atoms, **hp_config)

                    if hp_config['save_trajectory']:
                        write(os.path.join(out_dir, f'CONTCAR_{ads}_ads_{hp_config["calculation_index"]}_{i}.gat'),
                              atoms_ads)

                    energy_ads_list.append(energy_ads)

                    if force_max_ads > hp_config['fmax']:
                        converge_stat.append(0.0)
                    else:
                        converge_stat.append(1.0)

                energy_surf_list = np.array([energy_surf] * len(energy_ads_list))
                energy_ads_list = np.array(energy_ads_list)

                out = np.vstack([energy_ads_list, energy_surf_list, converge_stat]).T
                np.savetxt(f'ads_surf_energy_{ads}_{hp_config["calculation_index"]}.txt',
                           out, fmt='%f')

    def run(self, formula, **kwargs):
        # model save path
        energy_model_save_dir = self.hp_config['energy_model_save_dir']
        force_model_save_dir  = self.hp_config['force_model_save_dir']

        # instantiate a calculator
        calculator=GatAseCalculator(energy_model_save_dir, force_model_save_dir,
                                    self.hp_config['graph_build_scheme_dir'],
                                    gpu=self.hp_config['gpu'])

        with open('high_throughput_config.json', 'w') as f:
            json.dump(self.hp_config, f, indent=4)

        self.ads_calc(formula, calculator, **kwargs)

if __name__ == '__main__':
    HA = HpAds()
    HA.run('NiCoFePdPt', fmax=1.0, steps=2, gpu=0) # debug only



# Distributed under the terms of the MIT License.

__author__ = "Shyue Ping Ong, Geoffroy Hautier, Sai Jayaraman"
__copyright__ = "Copyright 2011, The Materials Project"

# =============================================================================
# Original author: "Shyue Ping Ong, Geoffroy Hautier, Sai Jayaraman"
# ZHANG Jun modified this script on 2021/05/14 (yyyy/mm/dd).
# Detail: get_connections() are replaced by get_connections_new(). The new one
# has a faster speed and more suitable for constructing a DGL graph.
# =============================================================================

from warnings import warn
import numpy as np
from scipy.spatial import Voronoi
from pymatgen.core.sites import PeriodicSite
from math import acos, pi
from pymatgen.util.num import abs_cap

class VoronoiConnectivity:
    """
    Computes the solid angles swept out by the shared face of the voronoi
    polyhedron between two sites.
    """

    def __init__(self, structure, cutoff=10):
        """
        Args:
            structure (Structure): Input structure
            cutoff (float) Cutoff distance.
        """
        self.cutoff = cutoff
        self.s = structure
        recp_len = np.array(self.s.lattice.reciprocal_lattice.abc)
        i = np.ceil(cutoff * recp_len / (2 * pi))
        offsets = np.mgrid[-i[0] : i[0] + 1, -i[1] : i[1] + 1, -i[2] : i[2] + 1].T
        self.offsets = np.reshape(offsets, (-1, 3))
        # shape = [image, axis]
        self.cart_offsets = self.s.lattice.get_cartesian_coords(self.offsets)

    @property
    def connectivity_array(self):
        """
        Provides connectivity array.

        Returns:
            connectivity: An array of shape [atomi, atomj, imagej]. atomi is
            the index of the atom in the input structure. Since the second
            atom can be outside of the unit cell, it must be described
            by both an atom index and an image index. Array data is the
            solid angle of polygon between atomi and imagej of atomj
        """
        # shape = [site, axis]
        cart_coords = np.array(self.s.cart_coords)
        # shape = [site, image, axis]
        all_sites = cart_coords[:, None, :] + self.cart_offsets[None, :, :]
        vt = Voronoi(all_sites.reshape((-1, 3)))
        n_images = all_sites.shape[1]
        cs = (len(self.s), len(self.s), len(self.cart_offsets))
        connectivity = np.zeros(cs)
        vts = np.array(vt.vertices)
        for (ki, kj), v in vt.ridge_dict.items():
            atomi = ki // n_images
            atomj = kj // n_images

            imagei = ki % n_images
            imagej = kj % n_images

            if imagei != n_images // 2 and imagej != n_images // 2:
                continue

            if imagei == n_images // 2:
                # atomi is in original cell
                val = solid_angle(vt.points[ki], vts[v])
                connectivity[atomi, atomj, imagej] = val

            if imagej == n_images // 2:
                # atomj is in original cell
                val = solid_angle(vt.points[kj], vts[v])
                connectivity[atomj, atomi, imagei] = val

            if -10.101 in vts[v]:
                warn("Found connectivity with infinite vertex. " "Cutoff is too low, and results may be " "incorrect")
        return connectivity

    @property
    def max_connectivity(self):
        """
        returns the 2d array [sitei, sitej] that represents
        the maximum connectivity of site i to any periodic
        image of site j
        """
        return np.max(self.connectivity_array, axis=2)

    def get_connections_new(self):     # modified by ZHANG Jun on 20210421, to build dgl graph
        """
        Returns a list of site pairs that are Voronoi Neighbors, along
        with their real-space distances.

        After test, this function returns the correct connections for a
        supercell big enough.
        """
        sender, receiver, dist = [], [], []
        maxconn = self.max_connectivity
        for ii in range(0, maxconn.shape[0]):
            for jj in range(ii + 1, maxconn.shape[1]):
                if maxconn[ii][jj] != 0:
                    dist.append(self.s.get_distance(ii, jj))
                    sender.append(ii)
                    receiver.append(jj)
        # print(np.shape(sender), np.shape(receiver), np.shape(dist))
        bsender   = sender   + receiver      # bidirectional
        breceiver = receiver + sender        # bidirectional
        bsender  += [x for x in range(maxconn.shape[0])] # add self_loop
        breceiver+= [x for x in range(maxconn.shape[0])] # add self_loop
        dist     *= 2                        # bidirectional
        dist     += [0.0] * maxconn.shape[0] # add self loop for `dist`
        # print(np.shape(bsender), np.shape(breceiver), np.shape(dist))
        dist      = [[x] for x in dist]
        # https://docs.dgl.ai/guide_cn/graph-feature.html 通过张量分配创建特征时，DGL会将特征赋给图中的每个节点和每条边。该张量的第一维必须与图中节点或边的数量一致。 不能将特征赋给图中节点或边的子集。
        return bsender, breceiver, dist

    def get_sitej(self, site_index, image_index):
        """
        Assuming there is some value in the connectivity array at indices
        (1, 3, 12). sitei can be obtained directly from the input structure
        (structure[1]). sitej can be obtained by passing 3, 12 to this function

        Args:
            site_index (int): index of the site (3 in the example)
            image_index (int): index of the image (12 in the example)
        """
        atoms_n_occu = self.s[site_index].species
        lattice = self.s.lattice
        coords = self.s[site_index].frac_coords + self.offsets[image_index]
        return PeriodicSite(atoms_n_occu, coords, lattice)

def solid_angle(center, coords):
    """
    Helper method to calculate the solid angle of a set of coords from the
    center.
    Args:
        center (3x1 array): Center to measure solid angle from.
        coords (Nx3 array): List of coords to determine solid angle.
    Returns:
        The solid angle.
    """
    o = np.array(center)
    r = [np.array(c) - o for c in coords]
    r.append(r[0])
    n = [np.cross(r[i + 1], r[i]) for i in range(len(r) - 1)]
    n.append(np.cross(r[1], r[0]))
    vals = []
    for i in range(len(n) - 1):
        v = -np.dot(n[i], n[i + 1]) / (np.linalg.norm(n[i]) * np.linalg.norm(n[i + 1]))
        vals.append(acos(abs_cap(v)))
    phi = sum(vals)
    return phi + (3 - len(r)) * pi
